import {ComboBox, combobox, EntityID, fieldset, NumberField, numberfield, t} from "@intermesh/goui";
import {client, FormWindow, jmapds} from "@intermesh/groupoffice-core";
import {ActivityCombo, activitycombo} from "@intermesh/business/business";
import {CalendarItem} from "@intermesh/community/calendar";

export class TimeEntryDialog extends FormWindow {
	private projectCombo: ComboBox;
	private readonly taskCombo: ComboBox;
	private activityCombo: ActivityCombo;
	private travelNumberField: NumberField;

	employeeId!: EntityID;
	eventId?: string | null;

	constructor() {
		super("Registration");

		this.stateId = "time-entry-dialog";
		this.title = t("Create time entry", "business", "calendartimetracking");
		this.resizable = true;
		this.height = 650;
		this.width = 400;

		this.generalTab.items.add(
			fieldset({},
				this.projectCombo = combobox({
					label: t("Project", "business", "projects3"),
					name: "projectId",
					valueProperty: "id",
					displayProperty: "name",
					placeholder: t("None"),
					clearable: true,
					required: true,
					dataSource: jmapds("Project3"),
					filter: {
						employeeId: this.employeeId
					},
					storeConfig: {
						filters: {
							default: {
								startsAt: "<= today",
								completedAt: null
							}
						}
					},
					listeners: {
						change: () => {
							this.taskCombo.reset();
						},
						setvalue: ( {newValue}) => {
							if (this.taskCombo) {
								this.taskCombo["value"] = undefined;
								this.taskCombo.filter!.projectId = newValue;
								this.taskCombo["disabled"] = false;
							}

							if (!newValue) {
								this.taskCombo["disabled"] = true;
								return;
							}
						}
					}
				}),
				this.taskCombo = combobox({
					label: t("Task", "community", "tasks"),
					name: "taskId",
					valueProperty: "id",
					displayProperty: "title",
					placeholder: t("Please select..."),
					disabled: true,
					dataSource: jmapds("Task"),
					filterName: "text",
					filter: {
						taskListId: "any",
						projectId: this.projectCombo.value
					}
				}),
				this.activityCombo = activitycombo({
					name: "activityId",
					required: true,
					storeConfig: {
						filters: {
							default: {
								type: "work"
							}
						}
					}
				}),
				this.travelNumberField = numberfield({
					label: t("Travel distance", "business", "projects3"),
					name: "distance",
					decimals: 2
				})
			)
		)

		this.form.on("beforesave", async ( {data}) => {
			const activity = await jmapds("Activity").single(data.activityId)

			if (activity && activity.billable !== 0) {
				if (activity.billable === 2) {
					data.billable = true;
				} else {
					const project = await jmapds("Project3").single(data.projectId);

					if (project && project.timeIsBillable) {
						data.billable = true;
					}
				}
			}
		});

		this.form.on("save", ( {data}) => {
			if (this.eventId !== null && data.id !== null) {
				const registrationId: number = Number(data.id);

				jmapds("CalendarEvent").update(this.eventId!, {registrationId: registrationId});
			}
		});
	}

	public loadCalItem(calendarItem: CalendarItem) {
		const startDate = new Date(calendarItem.initStart);
		const endDate = new Date(calendarItem.initEnd);
		const startTime =
			`${startDate.getHours().toString().padStart(2, "0")}:${startDate.getMinutes().toString().padStart(2, "0")}`;
		const comment = calendarItem.title;
		const duration = (endDate.getTime() - startDate.getTime()) / (1000 * 60);

		this.employeeId = calendarItem.cal.ownerId !== null ? calendarItem.cal.ownerId : client.user.id;

		this.projectCombo.filter!.isResource = this.employeeId;

		jmapds("Link").query({
			filter: {
				entity: "CalendarEvent",
				entityId: calendarItem.data.id
			},
			sort: [{property: "createdAt", isAscending: false}]
		}).then(r => {
			if (!r.ids.length)
				return

			jmapds("Link").get(r.ids).then(l => {
				if (!l)
					return

				for (let i = 0; i < l.list.length; i++) {
					if (l.list[i].toEntity === "Project3") {
						this.projectCombo.value = l.list[i].toId;
						return
					}
				}
			});
		});

		this.form.value = {
			employeeId: this.employeeId,
			startDate: startDate,
			startTime: startTime,
			bookedMinutes: duration,
			comment: comment
		}

		this.eventId = calendarItem.key;
	}
}