<?php /** @noinspection PhpUndefinedFieldInspection */

namespace go\modules\business\projects3\install;

use GO\Base\Db\FindParams;
use go\core\db\DbException;
use go\core\db\Query;
use go\core\db\Utils;
use go\core\jmap\Entity;
use go\core\model\Acl;
use go\core\model\FieldSet;
use go\core\model\Link;
use go\core\model\User;
use go\core\orm\exception\SaveException;
use go\core\util\Color;
use go\core\util\DateTime;
use go\modules\business\business\model\Activity;
use go\modules\business\business\model\Employee;
use go\modules\business\catalog\model\Article;
use go\modules\business\finance\model\FinanceDocumentPayment;
use go\modules\business\projects3\model\Project3;
use go\modules\business\projects3\model\ProjectBook;
use go\modules\business\projects3\model\ProjectMilestone;
use go\modules\business\registration\model\Registration;
use go\modules\community\history\Module;
use go\modules\community\tasks\model\TaskList;
use GO\Projects2\Model\Project;
use GO\Projects2\Model\Status;
use GO\Projects2\Model\TimeEntry;
use GO\Projects2\Model\Type;

class MigrateProjects2 {

	private int $businessId;
	public function run(int $businessId) {


		if(!\go\modules\business\projects3\Module::get()->isInstalled()) {
			if(!\go\modules\business\projects3\Module::get()->install()) {
				throw new \Exception("Could not install projects3");
			}
		}

		if(\go\core\model\Module::isInstalled("legacy", "timeregistration2") && !\go\modules\business\registration\Module::get()->isInstalled()) {
			if(!\go\modules\business\registration\Module::get()->install()) {
				throw new \Exception("Could not install registration");
			}
		}

		// Speed things up.
		Entity::$trackChanges = false;
		Module::$enabled = false;
		Entity::$checkFilesFolder = false;
		Project3::$updateSearch = false;

		$this->businessId = $businessId;

		FieldSet::migrateCustomFields(Project::class, Project3::class);

		$types = Type::model()->find();
		foreach($types as $type) {
			$this->migrateType($type);
		}

		if (\go\core\model\Module::isInstalled("community", "tasks")) {
			$this->migrateJobs();

		}


		echo "Done!\n";
	}

	/**
	 * @throws DbException
	 */
	public function reset(): void
	{
		go()->getDbConnection()->exec("delete from business_projects3_project3_custom_fields;

delete from business_projects3_project3_resource;

delete from business_registration_registration_custom_fields;

delete from business_registration_registration;

delete from business_projects3_project3;

delete from business_projects3_book_status;

delete from business_projects3_book;
");
	}


	private function getStatuses(): array {
		$old = Status::model()->find(FindParams::newInstance()->order('sort_order', 'ASC'));

		$new = [];
		foreach ($old as $status) {
			$new[] = [
				"name" => $status->name,
				"color" => trim(Color::background(),'#')
			];
		}

		return $new;
	}

	private function migrateType(Type $type): void
	{
		echo "Migrating type " . $type->name ."\n";

		$book = ProjectBook::findOrCreate($type->id, null, [
			"name" => $type->name,
			"businessId" => $this->businessId,
			"statuses" => $this->getStatuses()
		]);

		$book->findAcl()->copyFrom($type->findAclId());

		$projects = Project::model()->findByAttribute('type_id', $type->id);

		foreach($projects as $project) {
			$this->migrateProject($project, $book);
		}

		echo "\n\n";
	}


	private function migrateProject(Project $project, ProjectBook $book)
	{

		$resources = [];
		foreach ($project->resources as $resource) {

			$user = User::findById($resource->user_id, ['id', 'employee']);
			if ($user) {

				if($user->employee == null) {
					$user->employee = (new Employee($user))->setValues(['businessId' => $this->businessId]);
					if(!$user->save()) {
						throw new SaveException($user);
					}
				}

				$resources[$resource->user_id] = [
					"budgetMinutes" => $resource->budgeted_units ? $resource->budgeted_units * 60 : null,
					"fee" => $resource->external_fee
				];
			}
		}

		$statusId = null;
		$completedAt = null;
		if ($project->status_id) {
			$oldStatus = Status::model()->findByPk($project->status_id);
			if ($oldStatus) {

				if ($oldStatus->complete) {
					$completedAt = new DateTime("@" . $project->mtime);
				}
				foreach ($book->statuses as $s) {
					if ($s->name == $oldStatus->name) {
						$statusId = $s->id;
					}
				}
			}
		}

		$createdAt = new DateTime('@' . $project->ctime);

		$filesFolderId = null;
		if($project->files_folder_id > 0 && go()->getModule("legacy", "files")) {
			$exists = go()->getDbConnection()
				->selectSingleValue('id')
				->from("fs_folders")
				->where('id', '=', $project->files_folder_id)
				->single();
			if($exists) {
				$filesFolderId = $project->files_folder_id;
			}
		}

		echo "Create project " . $project->id . "\n";
		$p3 = Project3::findOrCreate($project->id, null, [
			'id' => $project->id,
			'name' => $this->cleanName($createdAt, $project->id, $project->name),
			'bookId' => $book->id,

			'createdBy' => User::exists($project->user_id) ? $project->user_id : 1,
			'modifiedBy' => User::exists($project->muser_id) ? $project->muser_id : 1,
			'modifiedAt' => new DateTime('@' . $project->mtime),
			'createdAt' => $createdAt,

			'startsAt' => $project->start_time ? new DateTime('@' . $project->start_time) : null,
			'dueAt' => $project->due_time ? new DateTime('@' . $project->due_time) : null,
			'completedAt' => $completedAt,
			'filesFolderId' => $filesFolderId,
			'contactId' => $project->contact_id,
			'customerId' => $project->company_id,

			"resources" => $resources,

			"statusId" => $statusId,

			"timeIsBillable" => $project->income_type == Project::INCOME_POST_CALCULATION

		], true);

		echo "Copying custom fields...\n";
		//using asText will lookup the new option ID's for select and multiselect fields.
		$p3->getCustomFields(true)->setValues($project->getCustomFields(true)->toArray());

		echo "Copying links...\n";
		Link::copyTo($project, $p3, true);

		if(\go\core\model\Module::isInstalled("legacy", "timeregistration2")) {
			$this->migrateHours($p3, $project, $book);
		}

		echo ".";


	}

	private function cleanName(DateTime $createdAt, int $id, string $name) : string {
		return trim(str_replace($createdAt->format('Y') . "-" . $id, "", $name));
	}

	private function migrateJobs(): void
	{
//		go()->getDbConnection()
//			->update(
//				"tasks_task",
//				't.projectId = l.projectId',
//				(new Query())
//					->tableAlias("t")
//					->join("tasks_tasklist", "l", "t.taskListId = l.id")
//			)->execute();

		$lists = TaskList::find()->where('projectId', '!=', null);

		echo "jobs\n";

		foreach($lists as $list) {

			if($this->projectHasMilestones($list->projectId)) {
				$project = Project3::findById($list->projectId);
				$milestone = new ProjectMilestone($project);
				$milestone->name = $list->name;

				$project->milestones[] = $milestone;
				if(!$project->save()) {
					throw new SaveException($project);
				}

				$milestoneId = $milestone->id;
			} else {
				$milestoneId = null;
			}

			$tasks = \go\modules\community\tasks\model\Task::find()->where('tasklistId', '=', $list->id);

			foreach ($tasks as $task) {

				echo ".";

				$task->mileStoneId = $milestoneId;
				$task->projectId = $list->projectId;
				if(!$task->save()) {
					throw new SaveException($task);
				}
			}
		}

		go()->getDbConnection()->exec("update tasks_tasklist set role=1 where role=3;");

		echo "\n";

	}

	private function migrateHours(Project3 $project, Project $oldProject, ProjectBook $book)
	{
		echo "\n\nhours\n\n";

		$hours = TimeEntry::model()->findByAttribute('project_id', $project->id, FindParams::newInstance()->select("*"));

		foreach($hours as $hour) {
			echo ".";
			$this->createRegistration($project, $oldProject, $book, $hour);
		}
	}

	private function createRegistration(Project3 $project, Project $oldProject, ProjectBook $book, TimeEntry $hour)
	{
		if($hour->units <= 0) {
			return;
		}
		//don't use model because we don't want validation and alerts etc.
		go()->getDbConnection()->replace("business_registration_registration",
		[
			'id' => $hour->id,
			'employeeId' => User::exists($hour->user_id) ? $hour->user_id : 1,
			'taskId' => $hour->task_id ? $hour->task_id : null,
			'projectId' => $project->id,
			'activityId' => Activity::exists($hour->standard_task_id)  ? $hour->standard_task_id : $this->createDefaultActivity()->id,
			'invoiceId' => null, //we can't determine this. we use a non billable activity type.
			'status' => $hour->status,
			'startDate' => (new DateTime("@" . $hour->date)),
			'endDate' => (new DateTime("@" . $hour->date)),
			'startTime' => (new DateTime("@" . $hour->date))->format("H:i"),
			'comment' => $hour->comments,
			'bookedMinutes' => $hour->units * 60,
			'distance' => $hour->travel_distance,
			'hourlyRevenue' => $hour->external_fee,
			'hourlyCosts' => $hour->internal_fee,
			'billable' => $hour->external_fee > 0,
			'billed' => !empty($hour->invoice_id),

			'createdBy' => User::exists($hour->user_id) ? $hour->user_id : 1,
			'modifiedBy' => 1,
			'modifiedAt' => new DateTime('@' . $hour->mtime),
			'createdAt' => new DateTime('@' . $hour->ctime)
		])->execute();

	}

	private ?Activity $defaultActivity;

	private function createDefaultActivity() : Activity {

		if(!isset($this->defaultActivity)) {
			$this->defaultActivity = Activity::findOrCreate('Migrated hours', "name", [
				'type' => Activity::TYPE_WORK,
				'billable' => Activity::BILLABLE_NO,
				"code" => "0000",
				"units" => 0,
				"disabled" => true
			]);
		}
		return $this->defaultActivity;
	}

	private function projectHasMilestones(int $projectId): bool
	{
		$count = go()->getDbConnection()
			->selectSingleValue('count(*)')
			->from('tasks_tasklist')
			->where('projectId', '=', $projectId)
			->single();

		return $count > 1;
	}

}