import {
	avatar,
	btn,
	column, comp,
	Component,
	DateInterval,
	EntityID,
	h3,
	Store,
	store,
	StoreRecord,
	t,
	Table,
	table,
	tbar
} from "@intermesh/goui";
import {client, jmapds, modules, router} from "@intermesh/groupoffice-core";
import {RegistrationWindow} from "@intermesh/business/registration";

export class ProjectResourceDetail extends Component {
	private table: Table<Store<StoreRecord>>;
	private projectId?: EntityID;
	constructor() {
		super();
		this.cls = "card vbox";


		this.hidden = true;

		this.table = this.createTable();

		this.items.add(
			tbar({}, h3("Resources"), "->", btn({
				icon: "expand_less",
				handler: (btn) => {
					this.table.hidden = !this.table.hidden;

					btn.icon = this.table.hidden ? "expand_more" : "expand_less";
				}
			})),
			comp({cls: "scroll", flex: 1},
				this.table
			)
		);

		jmapds("Registration").on("change", this.onRegistrationChange, {bind: this});

		this.on("remove", () => {
			jmapds("Registration").un("change", this.onRegistrationChange);
		})
	}


	private createTable() {

		const columns = [
			column({
				resizable: true,
				id: "principal",
				header: t("Resource"),
				renderer: (columnValue, record, td, table, storeIndex) => {

					if(columnValue) {

						avatar({
							cls: "inline",
							displayName: columnValue.name,
							backgroundImage: columnValue.avatarId ? client.downloadUrl(columnValue.avatarId) : undefined
						}).render(td);

						td.append(columnValue.name);


					} else {

						if(!record.bookedMinutes || !record.budgetMinutes) {
							return t("Total");
						}

						const perc = Math.ceil(record.bookedMinutes * 100 / record.budgetMinutes);

						return `${t("Total")} <progress max="100" value="${perc}">${perc}%</progress>`;
					}

				}
			}),

			column({
				width: 100,
				resizable: true,
				align: "right",
				id: "budgetMinutes",
				header: t("Budget"),
				renderer:(columnValue, record, td, table1) => {
					if(!columnValue) {
						return "-";
					}
					const di = DateInterval.createFromFormat(columnValue, "j");
					return di ? di.format("h:I") : "";
				}
			}),


			column({
				resizable: true,
				align: "right",
				id: "bookedMinutes",
				width: 100,
				header: t("Booked"),
				renderer:(columnValue, record, td, table1) => {
					if(!columnValue) {
						return "0:00";
					}
					const di = DateInterval.createFromFormat(columnValue, "j")!;

					return di.format("h:I");
				}
			}),

			column({
				resizable: true,
				align: "right",
				id: "percentage",
				property: "bookedMinutes",
				width: 100,
				header: t("Percentage"),
				renderer:(columnValue, record, td, table1) => {
					if(!columnValue || !record.budgetMinutes) {
						return "-";
					}

					const perc = Math.ceil(columnValue * 100 / record.budgetMinutes);
					return perc + "%";
				}
			}),

			column({
				width: 100,
				resizable: true,
				align: "right",
				id: "billableMinutes",
				header: t("Billable"),
				renderer:(columnValue, record, td, table1) => {
					const di = DateInterval.createFromFormat(columnValue, "j");
					return di ? di.format("h:I") : "";
				}
			}),



		];


		if(modules.isAvailable("business", "registration")) {
			columns.push(

				column({
					id: "add-time",
					width: 60,
					align: "right",
					sticky: true,
					renderer:(columnValue, record, td, table1, storeIndex) => {


						const btns = [];
						// happens for total line at bottom
						if(record.employeeId && (record.employeeId == client.user.id || modules.get("business", "registration")!.userRights.mayManage)) {
							btns.push(btn({
								icon: "add",
								title: t("Add time entry"),
								handler: async () => {
									const d = new RegistrationWindow();
									d.form.value = {
										employeeId: record.employeeId,
										projectId: this.projectId
									};
									const last = await d.lastRegistration();

									if(last)
										d.form.findField("activityId")!.value = last.activityId;


									d.show();
								}
							}))
						} else{
							btns.push(comp({flex: 1}));
						}

						btns.push(
							btn({
								icon: "timer",
								title: t("View time entries"),
								handler: () => {
									let url = "registration/list/" + this.projectId;

									if(record.employeeId) {
										url += "/" + record.employeeId;
									}

									router.goto(url);
								}
							})
						)

						// only allow manage users to track time for others

						return tbar({cls: "compact"},
							...btns
						)
					}
				})
			);
		}
		return table({
			fitParent: true,
			columns,
			store: store()
		})
	}

	private onRegistrationChange() {
		if(this.projectId) {
			void this.load(this.projectId);
		}
	}
	public async load(projectId: EntityID) {

		this.projectId = projectId;

		const response = await client.jmap("Project3/resources", {projectId: projectId});

		const resources = response.resources;

		if(!resources.length) {
			this.hide();
			return;
		} else {
			this.show();
		}

		await Promise.all(resources.map(async (r:any) => {
			r.principal = await jmapds("Principal").single(r.employeeId);
		}))

		const summary = resources.reduce((accumulator:any, currentValue:any) => {

			accumulator.bookedMinutes += currentValue.bookedMinutes
			accumulator.budgetMinutes += currentValue.budgetMinutes
			accumulator.billableMinutes += currentValue.billableMinutes

			return accumulator;
		}, {
			bookedMinutes: 0,
			budgetMinutes: 0,
			billableMinutes: 0,
		})

		resources.push(summary);

		this.table.store.loadData(resources, false);
	}
}