<?php

namespace go\modules\business\registration\install;

use DateTimeZone;
use Exception;
use go\core\db\Table;
use go\core\orm\Query;
use go\core\util\DateTime;
use go\core\util\StringUtil;
use go\modules\business\business\model\Activity;
use go\modules\business\registration\model\Registration;

class HoursMigrator
{
	/**
	 * Migrate the table pr2_hours to registrations_registration
	 *
	 * - Custom fields
	 * - For each user with registrations
	 * - clone the old record into the new one.
	 *
	 * @throws Exception
	 */
	public function hours2registrations()
	{
		$defaultActivityId = Activity::find(['type' => Activity::TYPE_WORK])->single()->id;
		echo "Migrating old time registration module" . PHP_EOL . PHP_EOL;

		go()->getCache()->flush();
		Table::destroyInstances();

		go()->getDbConnection()->beginTransaction();
		$db = go()->getDbConnection();

		// Migrate custom fields first
		$this->migrateCustomFields();

		// Split by user ID to reduce the huge number of chunkable records
		$users = $db->select('u.id, u.displayName, u.timezone')->from('core_user', 'u')
			->whereExists(((new Query())->select('ph.user_id'))->distinct()->from('pr2_hours', 'ph'))
			->join('pr2_hours', 'ph', 'u.id=ph.user_id')
			->distinct()
			->all();

		foreach ($users as $userRecord) {
			$count = 0;
			$userId = $userRecord['id'];
			$tz = new DateTimeZone($userRecord['timezone']);

			echo "Migrating time registrations for user " . $userRecord['displayName'] . " (" . $userId . ")" . PHP_EOL;
			flush();

			$hours = $db->select()->from('pr2_hours')
				->andWhere('user_id', '=', $userId)
				->orderBy(['date' => 'ASC']);
			foreach ($hours as $hour) {
				$hour = array_map("trim", $hour);
				echo ".";

				$count++;
				if ($count == 50) {
					echo "\n";
					$count = 0;
				}
				flush();
				$r = new Registration();
				$r->id = $hour['id'];
				$r->employeeId = $userId; // Remember: $userId === $employeeId
				$r->taskId = ($hour['task_id'] > 0) ? $hour['task_id'] : null;
				$r->projectId = $hour['project_id'];
				$r->activityId = $hour['standard_task_id'] ?? $defaultActivityId;
				$r->financeDocumentItemId = $hour['income_id'];
				$r->createdBy = 1; // As the admin is importing!
				$r->createdAt = DateTime::createFromFormat('U', $hour['ctime'])->setTimezone($tz);
				$r->modifiedBy = 1;
				$r->createdAt = DateTime::createFromFormat('U', $hour['mtime'])->setTimezone($tz);
				$r->distance = $hour['travel_distance'];
				//$r->costs = $hour['travel_costs']; -> This should be moved to finance!
				$r->status = $hour['status'];

				$regDate = DateTime::createFromFormat('U', $hour['date']);
				$regDate->setTimeZone($tz);
				$r->date = $regDate; //->format("Y-m-d");
				$r->startTime = $regDate->format("H:i:s");

				$r->bookedMinutes = round($hour['units'] * 60);
				$r->holidayUnits = 0;
				$r->comment = StringUtil::normalize($hour['comments']);
				$r->internalFee = $hour['internal_fee'];
				$r->externalFee = $hour['external_fee'];
				$r->billable = $r->externalFee > 0;
				$r->billed = !empty($hour['income_id']);
				$r->deleted = 0;
				$r->save();

			}
			echo "\n";
			flush();
		}

		go()->getDbConnection()->commit();
	}


	/**
	 * Migrate the custom fields. Should be easy
	 *
	 * If there are any field sets attached to the old pr2_hours entity, attach them to the new Registration entity
	 *
	 * @return void
	 */
	private function migrateCustomFields()
	{
		echo "Migrating custom fields" . PHP_EOL;
		flush();
		$c = go()->getDbConnection();
		$c->query("DROP TABLE IF EXISTS business_registration_registration_custom_fields");
		$c->query("CREATE TABLE business_registration_registration_custom_fields LIKE pr2_hours_custom_fields;");
		$c->query("INSERT INTO business_registration_registration_custom_fields SELECT * FROM pr2_hours_custom_fields;");

		echo "Cloning field sets and fields" . PHP_EOL;
		flush();
		
		$pr2HoursTable = Table::getInstance("pr2_hours_custom_fields");
		$cfTable = Table::getInstance("business_registration_registration_custom_fields");
		$cols = $pr2HoursTable->getColumns();
		$cols = array_filter($cols, function($n) {return $n->name != "model_id";});

		if(empty($cols)) {
			return;
		}

		$renameMap = [];
		foreach($cols as $col) {
			$i = 1;
			$name = $stripped = preg_replace('/\s+/', '_', $col->name);
			while($cfTable->hasColumn($name)) {
				$name = $stripped . '_' . $i++;
			}
			$renameMap[$col->name] = $name;
		}

		$timeEntryEntityType =  (new \go\core\db\Query)
			->select()
			->from('core_entity')
			->where('clientName = "TimeEntry"')
			->single();

		if($timeEntryEntityType) {

			foreach($renameMap as $old => $new) {
				go()->getDbConnection()
					->update("core_customfields_field",
						['databaseName' => $new],
						(new Query)
							->where(	'fieldSetId', 'IN',
								(new \go\core\db\Query)
									->select('id')
									->from('core_customfields_field_set')
									->where(['entityId' => $timeEntryEntityType['id']])
							)
							->andWhere('databaseName', '=', $old)
					)
					->execute();
			}

			go()->getDbConnection()
				->update("core_customfields_field_set",
					['entityId' => Registration::entityType()->getId()],
					['entityId' => $timeEntryEntityType['id']])
				->execute();
		}

		Table::destroyInstances();

	}
}