import {
	BaseEntity,
	comp,
	Component,
	datasourcestore,
	DataSourceStore,
	DateTime,
	Format, Store, StoreComponent,
	StoreRecord,
	t, Window
} from "@intermesh/goui";
import {JmapDataSource, jmapds} from "@intermesh/groupoffice-core";
import {getFirstMonday} from "./MainColumnView.js";
import {RegistrationEntity} from "./RegistrationEntity.js";
import {ColumnViewDay} from "./ColumnViewDay.js";

export type ViewMode = "week" | "month";

// TODO: This should probably be moved into the business module
export interface AgreementRecord extends BaseEntity {
	employeeId: number,
	start: DateTime,
	finish?: DateTime,
	mo: number,
	tu: number,
	we: number,
	th: number,
	fr: number,
	sa: number,
	su: number
}


interface LeavedayRecord extends StoreRecord {
	startDate: string,
	status: number,
	endDate: string,
	description: string,
	id: string
}

interface HolidayRecord extends StoreRecord {
	date: string,
	name: string,
	id: string
}


export class ColumnView extends Component implements StoreComponent<DataSourceStore<JmapDataSource<RegistrationEntity>>, RegistrationEntity> {
	public startDate: DateTime;
	private _employeeId: number | undefined;

	private dayCount: number | undefined;

	isClosed: boolean | undefined;
	closeDate?: DateTime | null;
	currentAgreement?: AgreementRecord;
	isManager?: boolean;

	store: DataSourceStore<JmapDataSource<RegistrationEntity>>;
	// holidayStore: Store<HolidayRecord>;
	// leavedayStore: Store<LeavedayRecord>;

	private tblCmp: Component;
	private totalDisplay?: Component;
	private days: ColumnViewDay[] = [];

	constructor(startDate: DateTime, viewMode: ViewMode = "week", employeeId?: number) {
		super();
		// this.id = "timeregColumnView";
		this.cls = "scroll";
		this.flex = 1;
		this.startDate = startDate;
		if (employeeId) {
			this._employeeId = employeeId;
		}
		this.store = datasourcestore({
			dataSource: jmapds("Registration"),
			relations: {
				project: {
					dataSource: jmapds("Project3"),
					path: "projectId"
				}
			},

			filters: {
				default: {
					isWork: 1,
					employeeId: this._employeeId
				}
			},
			sort: [
				{property: "startDate", isAscending: true},
				{property: "startTime", isAscending: true}
			]
		});

		this.store.bindComponent(this);

		// this.leavedayStore = store<LeavedayRecord>({
		// 	sort: [{property: "weeknb", isAscending: false}]
		// });
		//
		// this.holidayStore = store<HolidayRecord>({
		// 	sort: [{property: "weeknb", isAscending: false}]
		// });

		this.items.add(
			this.tblCmp = comp({tagName: "table", cls: "timeEntryTable card"})
		);

		this.setPeriod(this.startDate, viewMode);
	}

	public onStoreLoadException(ev:any) {
		void Window.error(ev.reason);
		this.unmask();
	}


	public onBeforeStoreLoad() {
		this.mask()
	}

	public onStoreLoad(ev:any) {
		this.drawView(this.startDate, ev.records);
		this.unmask();
	}

	public onRecordRemove() {

	}

	public onRecordAdd() {

	}

	public set employeeId(employeeId:number|undefined) {
		this._employeeId = employeeId;

		this.days.forEach(d => d.employeeId = employeeId)

	}

	public get employeeId() {
		return this._employeeId;
	}


	/**
	 * Draw the calendar view. Starts at the first date of the current period and will iterate until the last date
	 *
	 * Time registrations, leave days, weekends and holidays are all rendered
	 *
	 * @param currentDate
	 * @param timeEntries
	 * @private
	 */
	private drawView(currentDate: DateTime, timeEntries: RegistrationEntity[]) {
		this.tblCmp.items.clear();

		let dayComponents = [],
			headerComponents = [],
			periodTotal = 0;

		this.days = [];

		const shortWeekDays = ["mo", "tu", "we", "th", "fr", "sa", "su"];
		for (let d = 0; d < this.dayCount!; d++) {
			// Retrieve and sort today's time registration entries
			let todaysEntries = timeEntries.filter((r: RegistrationEntity) => {
				const date = r.startDate;
				return r.startDate == currentDate.format("Y-m-d");
			});

			const day = new ColumnViewDay(currentDate, todaysEntries);

			this.days.push(day);

			//TODO???
			//this.isClosed = (!!this.closeDate && (currentDate.date.getTime() <= this.closeDate.date.getTime()));

			day.workTime = this.currentAgreement![shortWeekDays[currentDate.getWeekDay()] as keyof AgreementRecord] as number
			day.employeeId = this._employeeId!;

			dayComponents.push(day.renderBody());
			headerComponents.push(day.renderHeader())
			periodTotal += day.getTotalTime();
			currentDate = currentDate.addDays(1);
		}

		if (!this.totalDisplay) {
			this.totalDisplay = comp({
				cls: "go-grid-total",
				tagName: "div",
				listeners: {
					render: ({target}) => {
						target.el.addEventListener("click", () => {
							target.hide();
						})
					}
				}
			});
			this.items.add(this.totalDisplay);
		}
		this.totalDisplay.html = t("Total") + " " + Format.duration(periodTotal / 60);

		this.tblCmp.items.add(
			comp({tagName: "tr"}, ...headerComponents),
			comp({tagName: "tr"}, ...dayComponents)
		);
		if (this.dayCount === 7) {
			// this.tblCntr.style.overflowX = "hidden";
			this.tblCmp.style.tableLayout = "fixed";
		} else {
			// this.tblCntr.style.overflowX = "scroll";
			this.tblCmp.style.tableLayout = "auto";
		}
		currentDate.addDays((0 - this.dayCount!));
	}

	/**
	 * Set start & end dates and day counts based on the current date and the selected view mode
	 *
	 * @param start
	 * @param viewMode
	 */
	public setPeriod(d: DateTime, viewMode: ViewMode) {
		let f;
		const start = d.clone();
		start.setHours(0).setMinutes(0).setSeconds(0);
		if (viewMode === "week") {
			// debugger;
			let weeknb = start.getWeekOfYear() - 1;
			//check if we're in week 1 that starts in last year
			if(weeknb == 0 && start.clone().addDays(7).getYear() != start.getYear()) {
				//skip to next week or we'll get the wrong first monday,
				start.addDays(7);
			}
			getFirstMonday(start);
			start.addDays(7 * weeknb);
			this.startDate = start;
			this.dayCount = 7;
			f = start.clone();
			f.addDays((this.dayCount - 1));
		} else {
			start.setDate(1);
			this.startDate = start;
			f = start.clone();
			f.addMonths(1).addDays(-1);
			this.dayCount = start.diff(f).getTotalDays()! + 1;
		}

		this.store.setFilter( "period", {endDate: ">=" + start.format("Y-m-d"), startDate: "<=" + f.format("Y-m-d")});
	}
}