import {
	btn,
	comp,
	Component,
	DateTime,
	EntityID,
	Format,
	hr,
	menu,
	NewEntity,
	root,
	t,
	Window,
	Window as GouiWindow,
} from "@intermesh/goui";
import {RegistrationEntity} from "./RegistrationEntity.js";
import {RegistrationWindow} from "./RegistrationWindow.js";
import {jmapds} from "@intermesh/groupoffice-core";
import {statuses} from "./MainColumnView.js";
import {CopyRegistrationWindow} from "./CopyRegistrationWindow.js";

export class ColumnViewDay {
	private readonly registrationComponents: Component[];
	private lastEndTime?: DateTime;
	private previousEntry?: RegistrationEntity;
	private nextStartTime?: string;



	constructor(private date:DateTime, private entries:RegistrationEntity[]) {
		this.sortEntries();
		this.registrationComponents = this.renderRegistrationComponents();
	}

	public getCls() {
		const clz = [];
		if (this.isToday()) {
			clz.push("today");
		}

		if (!this.isClosed) {
			clz.push("open");
		}
		// saturday or sunday
		if (this.date.getWeekDay() > 4) {
			clz.push("weekend");
		} else {
			clz.push("bgbox");
		}

		return clz.join(" ");
	}

	private totalTime = 0;

	public workTime = 0;

	public isClosed = false;

	public employeeId?:number;

	public isManager = false;

	/**
	 * The increment in minutes when resizing a registration.
	 */
	public step = 15;

	private minHeight = 80;

	private hourlyHeight = 80;

	private sortEntries() {
		this.entries.sort((a: RegistrationEntity, b: RegistrationEntity) => {
			const aMin = Format.minutes(a.startTime), bMin = Format.minutes(b.startTime);
			if (aMin < bMin) {
				return -1;
			} else if (bMin > aMin) {
				return 1;
			}
			return 0;
		});
	}

	private renderRegistrationComponents() {
		const entryComponents = [];
		this.lastEndTime = undefined;

		this.nextStartTime = this.date.format('Y-m-d') + ' 08:00';
		this.previousEntry = undefined;

		for (let i = 0, l = this.entries.length; i < l; i++) {
			const currentEntry = this.entries[i],
				currStartTime = DateTime.createFromFormat(currentEntry.startTime, "H:i")!,
				duration = currentEntry.bookedMinutes,
				startAt = DateTime.createFromFormat(this.nextStartTime!, "Y-m-d H:i")!;

			if (this.lastEndTime !== undefined && currStartTime.compare(this.lastEndTime) !== 0) {
				entryComponents.push(btn(
					{
						cls: "x-btn outlined",
						icon: "add",
						dataSet: {
							startAt: startAt,
							bookedMinutes: this.lastEndTime!.diff(currStartTime).getTotalMinutes()
						},
						text: t("Insert Time", "business", "registration"),
						hidden: this.isClosed,
						handler: (btn) => {

							this.addNewRegistration({
								startTime: btn.dataSet.startAt.format('H:i'),
								bookedMinutes: btn.dataSet.bookedMinutes,
								startDate: btn.dataSet.startAt.format('Y-m-d'),
								projectId: this.previousEntry?.project?.id,
								activityId: this.previousEntry?.activityId,
								employeeId: this.employeeId!
							});
						}
					}));
			}

			this.lastEndTime = currStartTime.addMinutes(duration).clone();

			let nextEntry = undefined;
			if (i < (l - 1)) {
				nextEntry = this.entries[(i + 1)];
			}

			// Add Time Entry to components
			entryComponents.push(this.timeEntryBlock(currentEntry, nextEntry));
			this.totalTime += (currentEntry.bookedMinutes * 60); // totalTime should be calculated in seconds!

			this.nextStartTime = DateTime.createFromFormat(currentEntry.startDate + " " +  currentEntry.startTime, "Y-m-d H:i")!.addMinutes(duration).format("Y-m-d H:i");

			this.previousEntry = currentEntry;
		}

		entryComponents.push(this.renderAddTime());

		return entryComponents;
	}

	private addNewRegistration(record: Partial<RegistrationEntity>) {
		const d = new RegistrationWindow();
		delete record.project;
		d.form.value = record;
		d.show();

		return d;
	}

	private isToday() {
		return this.date.format("Ymd") == (new DateTime()).format("Ymd");
	}

	public renderHeader() {
		let htmlStr = this.isToday() ? '<strong>' + t('Today') + '</strong>' : this.date.format("l");
		htmlStr += ' <span>' + this.date.format("j") + '</span><br>';
		htmlStr += '<strong>' + Format.duration((this.totalTime / 60));

		if(this.workTime) {
			htmlStr += " / " + Format.duration(this.workTime ?? 0);
		}

		htmlStr+=	'</strong>&nbsp;';

		htmlStr += ' <span class="monthLbl">' + this.date.format("M") + "</span>";

		return comp({
				tagName: "th",
				cls: this.getCls()
			},
			comp({cls: "reg-header-component", html: htmlStr})
		);
	}

	public getTotalTime() {
		return this.totalTime;
	}

	public renderBody() {
		return comp({
				tagName: "td",
				id: "td_" + this.date.format("Y-m-d"),
				cls: this.getCls()
			}, comp({
				listeners: {
					render: ({target}) => {
						if (this.isClosed) {
							return;
						}
						const cEl = target.el;
						cEl.addEventListener("dragenter", (evt) => {
							this.onDragEnter(evt, target)
						});

						cEl.addEventListener("dragleave", (evt: DragEvent) => {
							this.onDragLeave(evt, target);
						});

						cEl.addEventListener("dragover", (evt) => {
							evt.preventDefault();
							evt.dataTransfer!.dropEffect = evt.altKey ? "copy" : "move";
						});

						cEl.addEventListener("drop", (evt) => {
							this.onDrop(evt, target);
						});
					}
				}
			}, ...this.registrationComponents)
		)
	}

	private onDragEnter(evt: DragEvent, cmp:Component)  {
		evt.preventDefault();
		evt.dataTransfer!.effectAllowed = evt.altKey ? "copy" : "move";
		cmp.el.classList.add("drag-over");
	}

	private onDragLeave(_evt:DragEvent, cmp: Component) {
		cmp.el.classList.remove("drag-over");
	}
	private async onDrop(evt:DragEvent, _cmp:Component) {
		evt.preventDefault();
		const evtData = JSON.parse(evt.dataTransfer!.getData("text"));
		const evtTarget = evt.target;
		let tgtDate = null, tgtEl = evtTarget as HTMLElement, id = tgtEl!.id;
		do {
			if (id && id.match(/^td_[0-9]{4}-[0-9]{2}-[0-9]{2}$/)) {
				tgtDate = tgtEl!.id.substring(3);
			} else {
				tgtEl = tgtEl!.parentNode as HTMLElement;
				if (tgtEl) {
					id = tgtEl.id;
				}
			}
		} while (!tgtDate && tgtEl);

		if (tgtDate && evtData.startDate != tgtDate) {

			try {
				if (evt.altKey) {
					const newEntry = await jmapds("Registration").create({
						startDate: tgtDate,
						startTime: evtData.startTime,
						employeeId: evtData.employeeId,
						projectId: evtData.projectId,
						activityId: evtData.activityId,
						taskId: evtData.taskId,
						bookedMinutes: evtData.bookedMinutes,
						distance: evtData.distance,
						customFields: evtData.customFields,
						comment: evtData.comment
					});
				} else {
					const newEntry = await this.updateRegistrationEntity(evtData.id, {startDate: tgtDate});
				}
			} catch (e:any) {
				void Window.error(e);
			}
		}
	}

	private renderAddTime() {
		return btn({
			cls: "x-btn outlined",
			icon: "add",
			hidden: (this.isClosed),
			text: t("Add Time", "business", "registration"),
			listeners: {
				render: ({target}) => {

					let newTimeEntry: Component | undefined;

					const onMouseUp  = (evt:MouseEvent) => {
							this.onResizeEnd(evt, newTimeEntry!);
							window.removeEventListener("mousemove", onMouseMove);
							target.show();
						},
						onMouseMove = (e:MouseEvent) => {
							this.onResizing(e, newTimeEntry!);
						}

					target.el.addEventListener("mousedown", (e) => {

						const startAt = DateTime.createFromFormat(this.nextStartTime!, "Y-m-d H:i")!
						const reg = {
							startTime: startAt!.format("H:i"),
							bookedMinutes: 30,
							startDate: startAt!.format("Y-m-d"),
							projectId: this.previousEntry?.project?.id,
							project: this.previousEntry?.project,
							activityId: this.previousEntry?.activityId,
							employeeId: this.employeeId!
						};

						newTimeEntry = this.timeEntryBlock(reg);
						target.parent?.items.add(newTimeEntry);
						this.initResize(e, newTimeEntry);

						window.addEventListener("mousemove", onMouseMove);
						window.addEventListener("mouseup", onMouseUp, {once: true})


						target.hide();
					});
				}
			},

		});
	}


	/**
	 * Generate a time entry block based on the current and next time entry records
	 * These are to be rendered in the day column in the table.
	 *
	 * @param timeEntry
	 * @param nextEntry
	 * @private
	 */
	private timeEntryBlock(timeEntry: NewEntity<RegistrationEntity>, nextEntry?: RegistrationEntity) {
		let cls = ['time'];
		if (timeEntry.status == 2) {
			cls.push('disapproved');
		} else if (timeEntry.status == 0) {
			cls.push('open');
		}

		let maxEnd = 1440; // 24 hours
		if (nextEntry && nextEntry.startDate == timeEntry.startDate) {
			maxEnd = parseInt(nextEntry.startTime);
		}

		const cnt = this.getEntryContent(timeEntry);
		return comp({
			cls: cls.join(" "),
			dataSet: {timeEntry: timeEntry, maxEnd: maxEnd},
			height: Math.max(this.minHeight, (timeEntry.bookedMinutes! / 60) * this.hourlyHeight) - 8, //substract 8 for bottom margin on time entry
			listeners: {
				render: ({target}) => {
					const cEl = target.el;
					cEl.addEventListener("dblclick", (_evt) => {
						const d = new RegistrationWindow();
						d.show();
						void d.load(target.dataSet.timeEntry!.id);
					});
					cEl.addEventListener("contextmenu", (evt) => {

						evt.preventDefault();
						if (!this.isClosed) {
							this.showContextMenu(evt, target);
						}
					});



					if (timeEntry.id) {
						//only drag existing entries
						cEl.draggable = true;

						cEl.addEventListener("dragstart", (evt) => {
							evt.dataTransfer!.clearData();
							evt.dataTransfer!.setData("text/plain", JSON.stringify(target.dataSet.timeEntry));
							if (evt.altKey) {
								evt.dataTransfer!.effectAllowed = "copy";
								evt.dataTransfer!.dropEffect = "copy";
							} else {
								evt.dataTransfer!.effectAllowed = "move";
							}
						});
					}
				}
			}
		}, ...cnt);
	}

	/**
	 * Generate the content for a day entry based on several parameters
	 *
	 * @param values
	 * @private
	 */
	private getEntryContent(values: NewEntity<RegistrationEntity>) {

		let items = [];
		let commentStr = "";
		if (values.bookedMinutes >= 70) {
			if (values.comment) {
				commentStr += Format.escapeHTML(values.comment) + "<br>";
			} else {
				commentStr += "";
			}
		}
		commentStr += "<span>";
		if (values.status === 2) {
			commentStr += "<i class=\"icon\">warning</i>";
		}

		if(values.project) {
			commentStr += Format.escapeHTML(values.project.number + ": " + values.project.name) + "</span>";
		}
		items.push(comp({
			tagName: "p",
			html: commentStr
		}));

		items.push(comp({
			itemId: "time-string",
			tagName: "sub",
			html: this.timeString(values)
		}));

		if(values.id) {
			items.push(this.renderResizer());
		}
		return items;
	}

	private renderResizer() {
		return comp({
			tagName: "dd",
			listeners: {
				render: ({target}) => {
					const resizer = target.el,
						timeEntryCmp = target.parent!;

					const onMouseUp  = (evt:MouseEvent) => {
							window.removeEventListener("mousemove", onMouseMove);
							this.onResizeEnd(evt, timeEntryCmp);
						},
						onMouseMove = (e:MouseEvent) => {
							this.onResizing(e, timeEntryCmp);
						}

					resizer.addEventListener("mousedown", (evt) => {
						this.initResize(evt, timeEntryCmp);

						window.addEventListener("mousemove", onMouseMove);
						window.addEventListener("mouseup", onMouseUp, {once: true})
					});
				}
			}
		})
	}

	private resizeStartY = 0;
	private resizerCurrentModulo:number|undefined;
	private resizerStartBookedMinutes?: number;

	private initResize(evt:DragEvent|MouseEvent, timeEntryCmp:Component) {
		evt.preventDefault();
		this.resizeStartY = evt.clientY;
		this.resizerCurrentModulo = undefined;
		this.resizerStartBookedMinutes = parseInt(timeEntryCmp.dataSet.timeEntry.bookedMinutes);

		timeEntryCmp.parent!.el.classList.add("resize");
	}

	private onResizing(evt:MouseEvent, timeEntryCmp:Component) {
		evt.preventDefault();
		const deltaY = Component.pxToRem(evt.clientY - this.resizeStartY),
			data:RegistrationEntity = timeEntryCmp.dataSet.timeEntry,
			stepHeight = (this.hourlyHeight / 60) * this.step;

		const modulo = Math.floor(deltaY / stepHeight);
		if (modulo === this.resizerCurrentModulo) {
			return;
		}
		data.bookedMinutes = Math.max(0, this.resizerStartBookedMinutes! + (modulo * this.step));
		timeEntryCmp.height = Math.max(this.minHeight, data.bookedMinutes * stepHeight / this.step ) - 8; //substract 8 for bottom-margin on container
		timeEntryCmp!.findChild("time-string")!.html =  this.timeString(data);
		this.resizerCurrentModulo = modulo;

	}

	private async updateRegistrationEntity(id:EntityID, data:Partial<RegistrationEntity>) {
		try {
			root.mask();
			return await jmapds<RegistrationEntity>("Registration").update(id, data);
		} catch(e:any) {
			void Window.error(e);
			throw e;
		} finally {
			root.unmask();
		}
	}

	private async onResizeEnd(evt:MouseEvent, timeEntryCmp:Component) {
		timeEntryCmp.parent!.el.classList.remove("resize");

		evt.preventDefault();
		const data : RegistrationEntity = timeEntryCmp.dataSet.timeEntry;
		if (data.id) { // just resizing an existing one
			void this.updateRegistrationEntity(data.id, {
				bookedMinutes: data.bookedMinutes
			});

		} else {
			this.addNewRegistration(timeEntryCmp.dataSet.timeEntry).on("close", () => {
				timeEntryCmp.remove();
			})

		}
	}

	private timeString(values: NewEntity<RegistrationEntity>): string {
		const startTime = DateTime.createFromFormat(values.startDate + " " + values.startTime,"Y-m-d H:i")!;
		const endTime = startTime.clone().addMinutes(values.bookedMinutes);

		let timeStr = Format.duration(values.bookedMinutes) + "<em>";
		if (values.startTime) {
			timeStr += startTime.format(Format.timeFormat) + " - " +
				endTime.format(Format.timeFormat);
		}
		timeStr += "</em>";
		return timeStr;
	}

	private showContextMenu(evt: MouseEvent, cmp: Component) {
		let btns = [];

		const id: EntityID = cmp.dataSet.timeEntry.id, data: RegistrationEntity = cmp.dataSet.timeEntry;

		if (this.isManager && !this.isClosed) {
			btns.push(btn({
					text: t("Approve"),
					icon: "thumb_up",
					hidden: data.status == statuses.Approved,
					handler: async () => {
						data.status = statuses.Approved;
						void this.updateRegistrationEntity(id, {id: id, status: statuses.Approved});
					}
				}),
				btn({
					text: t("Disapprove"),
					icon: "thumb_down",
					hidden: data.status == statuses.Rejected,
					handler: async () => {
						const confirmed = await GouiWindow.confirm(t("Are you sure that you wish to disapprove the selected entry?"));
						if (confirmed) {
							data.status = statuses.Rejected;
							this.updateRegistrationEntity(id, {id: id, status: statuses.Rejected}).then(() => {
								GouiWindow.alert(t("Time registration disapproved"));
							});
						}
					}
				}),
				hr()
			);
		}
		btns.push(btn({
				text: t("Copy time entry"),
				icon: "content_copy",
				handler: (_btn) => {
					data.status = 0;
					const d = new CopyRegistrationWindow(data);
					d.show();
				}
			}),
			hr(), btn({
				text: t("Delete"),
				icon: "delete",
				handler: async () => {
					const confirmed = await GouiWindow.confirm(t("Are you certain that you wish to delete the selected entry?"));
					if (confirmed) {
						jmapds("Registration").destroy(id);
					}
				}
			}));

		const m = menu({isDropdown:true}, ...btns);
		m.showAt({x: evt.clientX, y: evt.clientY});
	}

}
