import {client, FormWindow, jmapds} from "@intermesh/groupoffice-core";
import {
	checkbox,
	CheckboxField,
	ComboBox,
	combobox,
	comp,
	datefield,
	DateField,
	DateTime,
	durationfield,
	DurationField,
	fieldset,
	HiddenField,
	hiddenfield,
	NumberField,
	numberfield,
	t,
	textarea, TextAreaField,
	timefield,
	TimeField
} from "@intermesh/goui";
import {ActivityCombo, activitycombo, EmployeeCombo, employeecombo} from "@intermesh/business/business";
import {statuses} from "./MainColumnView.js";
import {RegistrationEntity} from "./RegistrationEntity.js";

export class RegistrationWindow extends FormWindow<RegistrationEntity> {

	employeeCombo: EmployeeCombo;
	selectActivity: ActivityCombo;
	dateField: DateField;
	startTimeField: TimeField;
	endTimeField: TimeField;
	durationField: DurationField;
	projectCombo: ComboBox;
	taskCombo: ComboBox;
	travelDistanceField: NumberField;
	statusField: HiddenField;

	isManager!: boolean;
	private billableCB: CheckboxField;
	private commentField: TextAreaField;

	constructor() {
		super("Registration");
		this.title = t("Time registration");
		this.stateId = "time-registration-dialog";
		this.maximizable = false;
		this.resizable = true;
		this.width = 850;
		this.height = 700;

		const now = new DateTime();

		this.employeeCombo = employeecombo({
			hidden: true,
			required: true,
			name: "employeeId",
			value: client.user.id,
			listeners: {
				setvalue:( {newValue}) => {
					// Only show projects where the current user is a resource of
					if(this.projectCombo) {
						this.projectCombo.list.store.setFilter("isResource", {
							isResource: newValue
						});
					}
				}
			}
		});

		this.generalTab.items.add(
			fieldset({},

				comp({cls: "flow"},
					this.projectCombo = combobox({
						minWidth: 200,
						flex: 1,
						dataSource: jmapds("Project3"),
						name: "projectId",
						filterName: "text",
						label: t("Project", "business", "projects3"),
						storeConfig: {
							filters: {
								isResource: {isResource: this.employeeCombo.value},
								def: {
									startsAt: "<= today",
									completedAt: null
								}
							}
						},

						required: true,
						listeners: {
							change: () => {
								this.taskCombo.reset();
							},
							setvalue: ( {newValue}  ) => {
								if(this.taskCombo) {
									this.taskCombo.value = undefined;
									this.taskCombo.filter!.projectId = newValue;
								}

								if(!newValue) {
									return;
								}
								if(!this.form.currentId) {
									jmapds("Project3").single(newValue).then(project => {

										if (project) {
											this.billableCB.value = project.timeIsBillable;
										}
									})
								}
							}
						}
					}),

					this.taskCombo = combobox({
						flex: 1,
						minWidth: 200,
						dataSource: jmapds("Task"),
						displayProperty: "title",
						name: "taskId",
						label: t("Task", "community", "tasks"),
						filterName: "text",
						filter: {
							taskListId: "any",
							projectId: this.projectCombo.value
						},
						listeners: {
							change:async ( {newValue}) => {
								// debugger;
								if(newValue && !this.commentField.value) {
									const task = await jmapds("Task").single(newValue);
									if(task) {
										this.commentField.value = task.title;
									}
								}
							}
						}
					})
				),

				comp({cls: "flow"},
					this.selectActivity = activitycombo({
						required: true,
						flex: 1,
						minWidth: 200,
						storeConfig: {
							filters: {
								default: {
									type: "work"
								}
							}
						},
					}),
					this.billableCB = checkbox({
						width: "auto",
						type: "switch",
						name: "billable",
						label: t("Billable")
					}),
				),

				this.statusField = hiddenfield({
					//todo is this field needed?
					// type: "hidden",
					hidden: true,
					name: "status",
					value: statuses.Open
				}),


				this.commentField = textarea({
					autoHeight: true,
					name: "comment",
					label: t("What did you do?"),
					required: true
				}),

				comp({cls: "flow"},

					this.dateField = datefield({
						name: "startDate",
						required: true,
						label: t("Date"),
						value: now.format("Y-m-d"),
						flex: 1,
					}),

					this.startTimeField = timefield({
						label: t("Start time"),
						name: "startTime",
						flex: 1,
						value: now.clone().addHours(-1).format("H:i")
					}),

					this.endTimeField = timefield({
						label: t("End time"),
						id: "endTime",
						flex: 1,
						value: now.format("H:i")
					}),

					this.durationField = durationfield({
						label: t("Duration"),
						name: "bookedMinutes",
						outputFormat: "j", //total minutes
						flex: 1,
						value: 60
					})
				),

				this.travelDistanceField = numberfield({
					name: "distance",
					label: t("Travel distance", "business", "projects2"),
				}),
				this.employeeCombo
			)
		);

		this.employeeCombo.picker.on("select", ({record}) => {
			if (!client.user) {
				return;
			}
			this.statusField.value = statuses.Open;

			// If a manager updates a registration for one of their employees, it is approved by default.
			if (record.id != client.user.id && this.isManager) {
				this.statusField.value = statuses.Approved;
			}
		});

		this.on("render", async () => {

			// Should use the employee combo
			// if (client.user.employee!.timeClosedUntil) {
			// 	this.dateField.min = client.user.employee!.timeClosedUntil ?? undefined;
			// }
			if (client.user.isAdmin || client.user.employee.subjects.length) {
				this.isManager = true;
				this.employeeCombo.show();
				if (this.employeeCombo.value !== client.user.id) {
					this.statusField.value = statuses.Approved;
				}

				this.employeeCombo.picker.on("select", () => {

					this.selectActivity.value = undefined
					this.taskCombo.value = undefined;
					this.projectCombo.value = undefined;
				});
			}
		});

		const updateDuration = () => {
			const start = DateTime.createFromFormat(this.startTimeField.value!, "H:i")!,
				end = DateTime.createFromFormat(this.endTimeField.value!, "H:i")!,
				diff = start.diff(end);

			this.durationField.value = diff.getTotalMinutes();
		}

		// If the start time changes, add the duration in minutes to it, set end time field to result
		this.startTimeField.on("change",updateDuration);
		// If the end time changes, update the duration by subtracting the end time from it
		this.endTimeField.on("change", updateDuration);

		// If duration value changes, update end time by adding new duration to start time
		this.durationField.on("setvalue", ( {newValue}) => {
			const start = this.startTimeField.getValueAsDateTime();
			if(start) {
				this.endTimeField.value = start.addMinutes(newValue).format("H:i")
			}
		});

		this.on("ready", async ( {currentId}) => {
			if(!currentId && this.projectCombo.isEmpty()) {
				const lastRegistration = await this.lastRegistration()

				if(lastRegistration) {
					this.form.findField("projectId")!.value = lastRegistration.projectId;
					this.form.findField("activityId")!.value = lastRegistration.activityId;
					this.form.trackReset();
				}
			}
		})

	}


	public async lastRegistration() {
		const ds = jmapds<RegistrationEntity>("Registration");
		const lastIdQuery = await ds.query({
			filter: {employeeId: this.employeeCombo.value},
			sort: [{property: "startDate", isAscending:false}],
			limit: 1
		})

		if(!lastIdQuery.ids.length) {
			return;
		}

		return ds.single(lastIdQuery.ids[0]);
	}

}

