/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Config, Observable, ObservableEventMap } from "../Observable.js";
import { Table } from "./Table.js";
import { Component } from "../Component.js";
import { CheckboxField } from "../form/index.js";
/**
 * Return HTML or component to render into the table cell. Can also be async.
 *
 * Make sure the HTML it returns is HTML encoded so it's not vulnerable for XSS attacks.
 *
 * To create a multi line cell you can use h3,h4 and h5. For example:
 *
 * ```
 * return `<h3>${record.name.htmlEncode()}</h3><h4>${record.description.htmlEncode()}</h4><h5>${Format.smartDateTime(record.modifiedAt)}</h5>`
 * ```
 *
 * You can also return a component  like a {@link Button} or {@link Toolbar}. On a toolbar the CSS class "compact" is recommended.
 */
export type TableColumnRenderer = (columnValue: any, record: any, td: HTMLTableCellElement, table: Table, storeIndex: number, column: TableColumn) => string | Promise<string> | Component | Promise<Component> | undefined;
export type HeaderRenderer = (col: TableColumn, headerEl: HTMLTableCellElement, table: Table) => string | Component;
export type align = "left" | "right" | "center";
export interface TableColumnEventMap extends ObservableEventMap {
    /**
     * Fires when a table column renders
     */
    render: {
        /**
         * The result of the column renderer function
         */
        result: string | Promise<string> | Component | Promise<Component> | undefined;
        /**
         * The store record
         */
        record: any;
        /**
         * The record index in the store
         */
        storeIndex: number;
        /**
         * The table's td element
         */
        td: HTMLTableCellElement;
    };
}
export declare class TableColumn<EventMap extends TableColumnEventMap = TableColumnEventMap> extends Observable<EventMap> {
    id: string;
    parent: Table | undefined;
    /**
     *
     * The column ID. Also used for 'property'
     */
    constructor(id: string);
    /**
     * Path to property. If not given then 'id' is used
     *
     * @see ObjectUtil.get()
     */
    property: string;
    /**
     * Header in the table
     */
    header?: string;
    /**
     * Renderer function for the display
     */
    renderer?: TableColumnRenderer;
    /**
     * Renderer function for the header
     */
    headerRenderer?: HeaderRenderer;
    /**
     * Make the column resizable by the user
     */
    resizable: boolean;
    /**
     * Make it sortable by the user
     */
    sortable: boolean;
    /**
     * Width in rem units
     */
    width?: number;
    /**
     * Text alignment
     */
    align: align;
    /**
     * Hide the column. It can be enabled by the user via the context menu.
     */
    hidden: boolean;
    /**
     * Enable this column in the enabled columns menu
     */
    hidable: boolean;
    /**
     * When rendered this is set to the DOM element.
     * It's used to update the header width
     */
    headerEl?: HTMLTableCellElement;
    /**
     * Add CSS classes
     */
    cls?: string;
    /**
     * Makes the column stick when scrolling horizontally
     *
     * All the consecutive sticky columns from the beginning stick to the left side. Others will stick to the right side.
     *
     * Note:  The sticky column needs a background color. It is currently assumed to be on a background with "bg-lowest"
     */
    sticky?: boolean;
}
type TableColumnConfig = Config<TableColumn> & {
    /**
     * The ID of the column which is also the default for the column 'property'
     */
    id: string;
};
/**
 * Create a table column
 *
 * @param config
 */
export declare const column: (config: TableColumnConfig) => TableColumn<TableColumnEventMap>;
export declare class DateTimeColumn extends TableColumn {
    renderer: (date: string) => any;
    align: align;
    width: number;
}
/**
 * Create a column showing date and time
 * @param config
 */
export declare const datetimecolumn: (config: TableColumnConfig) => DateTimeColumn;
export declare class DateColumn extends TableColumn {
    renderer: (date: string) => any;
    align: align;
    width: number;
}
/**
 * Create a column showing just a date
 *
 * @param config
 */
export declare const datecolumn: (config: TableColumnConfig) => DateColumn;
export declare class BoolColumn extends TableColumn {
    renderer: (v: string) => "" | "<i class=\"icon\">check</i>";
    align: align;
    width: number;
}
/**
 * Create a column showing a boolean value as check or empty
 *
 * @param config
 */
export declare const boolcolumn: (config: TableColumnConfig) => BoolColumn;
export interface CheckboxColumnEventMap extends TableColumnEventMap {
    /**
     * Fires when the checkbox is checked by the user
     */
    change: {
        checkbox: CheckboxField;
        checked: boolean;
        record: any;
        storeIndex: number;
    };
}
export declare class CheckboxColumn extends TableColumn<CheckboxColumnEventMap> {
    constructor(id: string);
    renderer: TableColumnRenderer;
}
type CheckboxColumnConfig = Config<CheckboxColumn> & {
    /**
     * The ID of the column which is also the default for the column 'property'
     */
    id: string;
};
/**
 * Create a checkbox column
 *
 * @param config
 */
export declare const checkboxcolumn: (config: CheckboxColumnConfig) => CheckboxColumn;
export declare class CheckboxSelectColumn extends TableColumn {
    constructor(id?: string);
    headerRenderer: HeaderRenderer;
    renderer: TableColumnRenderer;
}
export declare const checkboxselectcolumn: (config?: CheckboxColumnConfig) => CheckboxSelectColumn;
/**
 * Creates a menu button.
 *
 * All items will have a property dataSet.rowIndex and dataSet.table so you know which record has been clicked on.
 *
 * @param items
 */
export declare const menucolumn: (...items: Component[]) => TableColumn<TableColumnEventMap>;
export {};
//# sourceMappingURL=TableColumns.d.ts.map