import { BaseEntity, Component, EntityID, MaterialIcon, Window } from "@intermesh/goui";
import { Entity } from "./Entities.js";
export interface EntityFilter {
    name: string;
    type: string;
    typeConfig?: Record<string, any>;
    title: string;
    multiple: boolean;
}
export interface EntityLink {
    /**
     * Filter key. For contacts there's "isOrganization" for example.
     */
    filter?: string;
    /**
     * Human friendly title for menu's. Defaults to t(entity.name)
     */
    title?: string;
    /**
     * CSS class to render the icon
     */
    iconCls: string;
    /**
     * Use this link only for the search filter. Used for comments.
     */
    searchOnly?: boolean;
    /**
     * Create a window that will create a new linked item of this type
     *
     * @param entity
     * @param entityId
     */
    linkWindow?: (entity: string, entityId: EntityID) => Window;
    /**
     * Return a detail component to show a linked entity of this type
     */
    linkDetail: () => Component;
}
export interface EntityConfig {
    /**
     * Entity name
     *
     * eg. "Contact"
     */
    name: string;
    /**
     * Human friendly title for the entity
     *
     * If not given it will default to t(entity.name);
     */
    title?: string;
    /**
     * Available linking options.
     *
     * In most cases there will be one but for example a contact as one extra for organizations. The "filter" option is
     * used to differentiate the two.
     */
    links?: EntityLink[];
    /**
     * Custom filters for the entity
     */
    filters?: EntityFilter[];
}
export type ModuleConfig = {
    /**
     * Module package name
     */
    package: string;
    /**
     * Module name
     */
    name: string;
    /**
     * Init function. Is called when the main Group-Office page loads
     */
    init?: () => void;
    /**
     * Registered module entities
     */
    entities?: (string | EntityConfig)[];
};
declare global {
    var GO: any;
    var go: any;
    var Ext: any;
    var BaseHref: string;
}
export interface Module extends BaseEntity {
    name: string;
    package: string;
    rights: string[];
    settings?: Record<string, any>;
    userRights: Record<string, boolean>;
    version: number;
    entities: Record<string, Entity>;
}
declare class Modules {
    private mods;
    private modules?;
    /**
     * Register a module so it's functionally is added to the GUI
     *
     * @param config
     */
    register(config: ModuleConfig): void;
    /**
     * Add a main panel that is accessible through the main menu and tabs
     *
     * @param pkg
     * @param module
     * @param id
     * @param title
     * @param callback
     */
    addMainPanel(pkg: string, module: string, id: string, title: string, callback: () => Component | Promise<Component>): void;
    /**
     * Add a system settings panel
     *
     * @param pkg
     * @param module
     * @param id
     * @param title
     * @param icon
     * @param callback
     */
    addSystemSettingsPanel(pkg: string, module: string, id: string, title: string, icon: MaterialIcon, callback: () => Component | Promise<Component>): void;
    /**
     * Add a system settings panel
     *
     * @param pkg
     * @param module
     * @param id
     * @param title
     * @param icon
     * @param callback
     */
    addAccountSettingsPanel(pkg: string, module: string, id: string, title: string, icon: MaterialIcon, callback: () => Component): void;
    /**
     * Open a main panel
     *
     * @param id
     */
    openMainPanel(id: string): void;
    private registerInExtjs;
    /**
     * Get all modules
     */
    getAll(): Module[];
    /**
     * Check if the current user has this module
     *
     * @param pkg
     * @param name
     */
    isAvailable(pkg: string, name: string): boolean;
    /**
     * Get a module
     *
     * @param pkg
     * @param name
     */
    get(pkg: string, name: string): Module | undefined;
}
export declare const modules: Modules;
export {};
//# sourceMappingURL=Modules.d.ts.map