/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Michael de Hart <mdhart@intermesh.nl>
 */
import { Component, ComponentEventMap } from "./Component.js";
import { Store, StoreComponent, StoreRecord } from "../data";
import { RowSelect, RowSelectConfig } from "./table";
import { Config, Listener, ObservableListenerOpts } from "./Observable.js";
export type RowRenderer = (record: any, row: HTMLElement, list: any, storeIndex: number) => string | Component[] | void;
export type GroupByRenderer = (groupBy: any, record: any, list: List) => string | Promise<string> | Component | Promise<Component>;
export type listStoreType<ListType> = ListType extends List<infer StoreType> ? StoreType : never;
/**
 * @inheritDoc
 */
export interface ListEventMap<Type> extends ComponentEventMap<Type> {
    /**
     * Fires when the user scrolled to the bottom
     *
     * @param list
     */
    scrolleddown: (list: Type) => void;
    /**
     * Fires when the user sorts the list by a property
     *
     * @param list
     * @param property
     */
    sort: (list: Type, property: string) => void;
    /**
     * Fires when the user sorts the list by drag and drop
     *
     * @param list
     * @param dataIndex
     */
    sortchange: (list: Type, record: any, dropIndex: number, oldIndex: number) => void;
    /**
     * Fires when a row is mousedowned
     *
     * @param list
     * @param storeIndex
     * @param ev
     */
    rowmousedown: (list: Type, storeIndex: number, row: HTMLElement, ev: MouseEvent) => void;
    /**
     * Fires when a row is clicked
     *
     * @param list
     * @param storeIndex
     * @param ev
     */
    rowclick: (list: Type, storeIndex: number, row: HTMLElement, ev: MouseEvent | KeyboardEvent) => void;
    /**
     * Fires when a row is double clicked
     *
     * @param list
     * @param storeIndex
     * @param ev
     */
    rowdblclick: (list: Type, storeIndex: number, row: HTMLElement, ev: MouseEvent) => void;
    /**
     * Fires when the delete key is pressed
     *
     * @example
     * ```
     * delete: async (list) => {
     * 	const ids = list.rowSelection!.selected.map(index => list.store.get(index)!.id);
     * 	await jmapds("Foo").confirmDestroy(ids);
     * }
     * ```
     *
     * @param list
     */
    delete: (list: Type) => void;
    /**
     * Fires when a row is right clicked
     *
     * @param list
     * @param storeIndex
     * @param ev
     */
    rowcontextmenu: (list: Type, storeIndex: number, row: HTMLElement, ev: MouseEvent) => void;
    /**
     * Fires when records are rendered into rows.
     *
     * @param list
     * @param records
     */
    renderrows: (list: Type, records: any[]) => void;
    /**
     * Fires when a row is clicked or navigated with arrows
     *
     * @param list
     * @param storeIndex
     * @param record
     */
    navigate: (list: Type, storeIndex: number) => void;
    /**
     * Fires when something was dropped
     *
     * @param toComponent The component where it's dropped on. Usually this component but with trees it can be a nested list.
     * @param fromIndex The index in the fromComponent of the item being dragged
     * @param toIndex The index where it's dropped in this list
     * @param droppedOn True if dropped on a node and not between
     * @param fromComponent The component where the item is dragged from. When the same sort group is used it can be another component
     * @param dragDataSet Arbitrary drag data components may set. A list adds dragDataSet.selectedRowIndexes: The row indexes when a multiselect is dragged. If the record dragged is not part of the selection then it will contain the single dragged record.
     */
    drop: (toComponent: Type, toIndex: number, fromIndex: number, droppedOn: boolean, fromComp: Component, dragDataSet: Record<string, any>) => void;
    /**
     * Fires when the items are dragged over this list.
     *
     * Return false to disallow dropping
     *
     * @param dropComp The component the element was dropped on
     * @param fromIndex Move from index
     * @param toIndex To index
     * @param droppedOn Dropped on the toIndex or moved to this index
     * @param fromComp The component the element was dragged from if "group" is used to drop to other components
     * @param dragDataSet Arbitrary drag data components may set. A list adds dragDataSet.selectedRowIndexes: The row indexes when a multiselect is dragged. If the record dragged is not part of the selection then it will contain the single dragged record.
     */
    dropallowed: (toComponent: Type, toIndex: number, fromIndex: number, droppedOn: boolean, fromComp: Component, dragDataSet: Record<string, any>) => void;
}
export interface List<StoreType extends Store = Store> extends Component {
    on<K extends keyof ListEventMap<this>, L extends Listener>(eventName: K, listener: Partial<ListEventMap<this>>[K], options?: ObservableListenerOpts): L;
    un<K extends keyof ListEventMap<this>>(eventName: K, listener: Partial<ListEventMap<this>>[K]): boolean;
    fire<K extends keyof ListEventMap<this>>(eventName: K, ...args: Parameters<ListEventMap<any>[K]>): boolean;
}
/**
 * List component
 *
 * Create a list with a custom item renderer. Also capable of selecting rows.
 */
export declare class List<StoreType extends Store = Store> extends Component implements StoreComponent<StoreType> {
    readonly store: StoreType;
    readonly renderer: RowRenderer;
    /**
     * Shown when the list is empty.
     */
    emptyStateHtml: string;
    protected emptyStateTag: keyof HTMLElementTagNameMap;
    private emptyEl?;
    private rowSelect?;
    /**
     * Allow items to be dragged
     */
    draggable: boolean;
    /**
     * Allow to drop between items
     */
    dropBetween: boolean;
    /**
     * Allow to drop on items
     */
    dropOn: boolean;
    /**
     * Group for sortable when drag and drop is used
     */
    sortableGroup: string | undefined;
    /**
     * When enabled, it will register it's container as scrolling element to the {@link Store}.
     * @see Store.addScrollLoader();
     */
    scrollLoad: boolean;
    /**
     * Group the table by this property.
     */
    groupBy?: string;
    /**
     * Makes the groups collapsible
     */
    groupByCollapsible: boolean;
    /**
     * Group renderer function
     */
    groupByRenderer: GroupByRenderer;
    protected itemTag: keyof HTMLElementTagNameMap;
    private lastGroup?;
    /**
     * Row selection object
     * @param rowSelectionConfig
     */
    set rowSelectionConfig(rowSelectionConfig: boolean | Partial<RowSelectConfig>);
    constructor(store: StoreType, renderer: RowRenderer, tagName?: keyof HTMLElementTagNameMap);
    get rowSelection(): RowSelect | undefined;
    get el(): HTMLElement;
    protected internalRender(): HTMLElement;
    protected initSortable(): void;
    protected onKeyDown(e: KeyboardEvent): void;
    protected initStore(): void;
    onStoreLoadException(store: StoreType, reason: any): void;
    onBeforeStoreLoad(store: StoreType, append: boolean): void;
    onStoreLoad(): void;
    onRecordRemove(collection: StoreType, item: any, index: number): void;
    onRecordAdd(collection: StoreType, item: StoreRecord, index: number): void;
    protected getRowElements(): HTMLElement[];
    private initNavigateEvent;
    protected renderEmptyState(): void;
    protected renderBody(): void;
    focusRow(index: number): boolean;
    getFocussedRow(): {
        rowEl: HTMLElement;
        index: number;
    } | undefined;
    protected rerender(): void;
    private groupEl;
    protected renderRows(records: any[]): void;
    protected isNewGroup(record: any): any;
    protected onRowAppend(row: HTMLElement, record: any, index: number): void;
    protected renderGroup(record: any): HTMLElement;
    private wrapGroupByCollapsible;
    private toggleGroup;
    protected renderGroupToEl(record: any, groupDisplayEl: HTMLElement): void;
    protected renderRow(record: any, storeIndex: number): HTMLElement;
    private onMouseEvent;
    private findRowByEvent;
}
export type ListConfig<StoreType extends Store> = Omit<Config<List<StoreType>, ListEventMap<List<StoreType>>, "store" | "renderer">, "rowSelection">;
/**
 * Shorthand function to create {@link Table}
 *
 * @param config
 */
export declare const list: <StoreType extends Store>(config: ListConfig<StoreType>) => List<StoreType>;
//# sourceMappingURL=List.d.ts.map