/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Michael de Hart <mdhart@intermesh.nl>
 */
import { assignComponentConfig, comp, Component, createComponent, span } from "./Component.js";
import { t } from "../Translate.js";
import { E, ObjectUtil } from "../util";
import { rowselect, SelectedRow } from "./table";
import { Window } from "./Window.js";
import { Sortable } from "./Sortable.js";
/**
 * List component
 *
 * Create a list with a custom item renderer. Also capable of selecting rows.
 */
export class List extends Component {
    /**
     * Row selection object
     * @param rowSelectionConfig
     */
    set rowSelectionConfig(rowSelectionConfig) {
        if (typeof rowSelectionConfig != "boolean") {
            if (!this.rowSelect) {
                rowSelectionConfig.list = this;
                this.rowSelect = rowselect(rowSelectionConfig);
            }
            else {
                assignComponentConfig(this.rowSelect, rowSelectionConfig);
            }
        }
        else if (!this.rowSelect) {
            this.rowSelect = rowselect({ list: this });
        }
    }
    constructor(store, renderer, tagName = "ul") {
        super(tagName);
        this.store = store;
        this.renderer = renderer;
        /**
         * Shown when the list is empty.
         */
        this.emptyStateHtml = `<div class="goui-empty-state"><i class="icon">article</i><p>${t("No items found")}</p></div>`;
        this.emptyStateTag = 'li';
        /**
         * Allow items to be dragged
         */
        this.draggable = false;
        /**
         * Allow to drop between items
         */
        this.dropBetween = false;
        /**
         * Allow to drop on items
         */
        this.dropOn = false;
        /**
         * Group for sortable when drag and drop is used
         */
        this.sortableGroup = undefined;
        /**
         * When enabled, it will register it's container as scrolling element to the {@link Store}.
         * @see Store.addScrollLoader();
         */
        this.scrollLoad = false;
        /**
         * Makes the groups collapsible
         */
        this.groupByCollapsible = true;
        /**
         * Group renderer function
         */
        this.groupByRenderer = groupBy => groupBy ?? t("None");
        this.itemTag = 'li';
        this.tabIndex = 0;
    }
    get rowSelection() {
        return this.rowSelect;
    }
    get el() {
        return super.el;
    }
    internalRender() {
        const el = super.internalRender();
        this.initNavigateEvent();
        el.on("mousedown", (e) => {
            this.onMouseEvent(e, "rowmousedown");
        }).on("dblclick", (e) => {
            this.onMouseEvent(e, "rowdblclick");
        }).on("click", (e) => {
            this.onMouseEvent(e, "rowclick");
        }).on("contextmenu", (e) => {
            this.onMouseEvent(e, "rowcontextmenu");
        }).on("keydown", (e) => {
            this.onKeyDown(e);
        });
        this.renderEmptyState();
        this.renderBody();
        this.initStore();
        this.initSortable();
        return el;
    }
    initSortable() {
        if (!this.dropBetween && !this.dropOn && !this.draggable) {
            return;
        }
        const sortable = new Sortable(this, ".data");
        sortable.dropOn = this.dropOn;
        sortable.dropBetween = this.dropBetween;
        if (!this.sortableGroup) {
            this.sortableGroup = "sortable-" + Component.uniqueID();
        }
        sortable.group = this.sortableGroup;
        sortable.on("sort", (toComp, toIndex, fromIndex, droppedOn, fromComp, dragDataSet) => {
            return this.fire("drop", toComp, toIndex, fromIndex, droppedOn, fromComp, dragDataSet);
        });
        sortable.on("dropallowed", (toComp, toIndex, fromIndex, droppedOn, fromComp, dragDataSet) => {
            return this.fire("dropallowed", this, toIndex, fromIndex, droppedOn, fromComp, dragDataSet);
        });
        sortable.on("dragstart", (sortable1, ev, dragData) => {
            //Multiselect movement support here. Don't move selection if the dragged element wasn't part of the selection
            if (this.rowSelect && this.rowSelect.multiSelect && this.rowSelect.getSelected().length > 1 && this.rowSelect.isSelected(this.store.get(dragData.fromIndex))) {
                dragData.dataSet.selectedRowIndexes = this.rowSelect.getSelected();
                ev.setDragComponent(comp({ cls: "card pad", html: this.rowSelect.getSelected().length + " selected rows" }));
            }
            else {
                dragData.dataSet.selectedRowIndexes = [new SelectedRow(this.store, this.store.get(dragData.fromIndex))];
            }
        });
        sortable.on("dragend", (sortable1, ev, dragData1) => {
            delete dragData1.dataSet.selectedRowIndexes;
        });
    }
    onKeyDown(e) {
        if (e.key == "Delete" || e.metaKey && e.key == "Backspace") {
            e.preventDefault();
            this.fire("delete", this);
        }
        if (e.key == "Enter") {
            const r = this.getFocussedRow();
            if (r) {
                e.preventDefault();
                this.fire("rowclick", this, r.index, r.rowEl, e);
            }
        }
    }
    initStore() {
        this.store.bindComponent(this);
        if (this.scrollLoad && this.parent) {
            this.store.addScrollLoader(this.parent.el);
        }
    }
    onStoreLoadException(store, reason) {
        void Window.error(reason);
        this.unmask();
    }
    onBeforeStoreLoad(store, append) {
        this.mask();
        if (this.store.loaded && !append) {
            // when loading (not reloading, then this.store.loaded is false) we want to reset the parent's scrolltop
            if (this.parent && this.parent.el.scrollTop) {
                this.parent.el.scrollTop = 0;
            }
        }
    }
    onStoreLoad() {
        this.unmask();
        if (this.emptyEl) {
            this.emptyEl.hidden = this.store.count() > 0;
        }
    }
    onRecordRemove(collection, item, index) {
        const rows = this.getRowElements();
        rows[index]?.remove();
        // Remove row from selection too if it's not caused by a store load. Then we want to maintain the selection.
        if (!this.store.loading && this.rowSelection) {
            this.rowSelection.remove(item, true);
        }
    }
    //todo inserting doesn't work with groups yet. It can only append to the last
    onRecordAdd(collection, item, index) {
        let container = this.groupEl || this.el;
        const newGroup = this.isNewGroup(item);
        if (newGroup !== false) {
            container = this.groupEl = this.renderGroup(item);
        }
        const row = this.renderRow(item, index);
        if (index == collection.count() - 1) {
            container.append(row);
        }
        else {
            const before = this.getRowElements()[index];
            before.parentElement.insertBefore(row, before);
        }
        this.onRowAppend(row, item, index);
    }
    getRowElements() {
        return Array.from(this.el.querySelectorAll(":scope > .data"));
    }
    initNavigateEvent() {
        this.on('rowmousedown', (list, storeIndex, row, ev) => {
            if (!ev.shiftKey && !ev.ctrlKey) {
                this.fire("navigate", this, storeIndex);
            }
        });
        if (this.rowSelection) {
            this.el.addEventListener('keydown', (ev) => {
                if (!ev.shiftKey && !ev.ctrlKey && (ev.key == "ArrowDown" || ev.key == "ArrowUp")) {
                    const selected = this.rowSelect.getSelected();
                    if (selected.length) {
                        this.fire("navigate", this, selected[0].storeIndex);
                    }
                }
            });
        }
    }
    renderEmptyState() {
        this.emptyEl = E(this.emptyStateTag).css({ 'captionSide': 'bottom', height: '100%' });
        this.emptyEl.hidden = this.store.count() > 0;
        this.emptyEl.innerHTML = this.emptyStateHtml;
        this.el.appendChild(this.emptyEl);
    }
    renderBody() {
        this.renderRows(this.store.all());
        if (this.rowSelect) {
            this.rowSelect.on('rowselect', (rowSelect, row) => {
                const tr = this.getRowElements()[row.storeIndex];
                if (!tr) {
                    //row not rendered (yet?). selected class will also be added on render
                    return;
                }
                tr.classList.add('selected');
                tr.scrollIntoView({ block: "nearest", inline: "nearest" });
            });
            this.rowSelect.on('rowdeselect', (rowSelect, row) => {
                const tr = this.getRowElements()[row.storeIndex];
                if (!tr) {
                    console.error("No row found for selected index: " + row.storeIndex + ". Maybe it's not rendered yet?");
                    return;
                }
                tr.classList.remove('selected');
            });
        }
    }
    focusRow(index) {
        const tr = this.getRowElements()[index];
        if (!tr) {
            return false;
        }
        tr.focus();
        return true;
    }
    getFocussedRow() {
        if (!document.activeElement) {
            return undefined;
        }
        const rows = this.getRowElements();
        const index = rows.indexOf(document.activeElement);
        if (index == -1) {
            return undefined;
        }
        return { rowEl: rows[index], index: index };
    }
    rerender() {
        const el = this.el;
        this.groupEl = undefined;
        this.lastGroup = undefined;
        el.innerHTML = "";
        this.renderBody();
    }
    renderRows(records) {
        let container = this.groupEl || this.el;
        for (let i = 0, l = records.length; i < l; i++) {
            const newGroup = this.isNewGroup(records[i]);
            if (newGroup !== false) {
                container = this.groupEl = this.renderGroup(records[i]);
            }
            const row = this.renderRow(records[i], i);
            container.append(row);
            this.onRowAppend(row, records[i], i);
        }
        this.fire("renderrows", this, records);
    }
    isNewGroup(record) {
        const groupBy = this.groupBy ? ObjectUtil.get(record, this.groupBy) : "";
        if (this.lastGroup != groupBy) {
            this.lastGroup = groupBy;
            return groupBy;
        }
        else {
            return false;
        }
    }
    onRowAppend(row, record, index) {
    }
    renderGroup(record) {
        // return this.fragment!;
        if (this.groupBy) {
            const li = document.createElement("li");
            li.classList.add("group");
            this.renderGroupToEl(record, li);
            this.el.append(li);
        }
        return this.el;
    }
    wrapGroupByCollapsible(groupDisplayEl) {
        // this component will replace the original
        const groupDisplayComp = comp();
        const groupRow = comp({
            cls: "hbox"
        }, span({
            itemId: "expander",
            cls: "icon",
            text: "expand_more"
        }), groupDisplayComp);
        groupRow.render(groupDisplayEl);
        groupRow.el.addEventListener("click", (ev) => {
            this.toggleGroup(groupRow);
        });
        return groupDisplayComp.el;
    }
    toggleGroup(groupRow) {
        const groupContainer = groupRow.el.closest("tbody");
        if (!groupContainer) {
            return;
        }
        const icon = groupRow.findChild("expander");
        icon.text == "expand_more" ? icon.text = "chevron_right" : icon.text = "expand_more";
        const dataRows = groupContainer.querySelectorAll("tr.data");
        dataRows.forEach((row) => {
            const tableRow = row;
            tableRow.style.display == "none" ? tableRow.style.display = "" : tableRow.style.display = "none";
        });
    }
    renderGroupToEl(record, groupDisplayEl) {
        if (this.groupByCollapsible) {
            groupDisplayEl = this.wrapGroupByCollapsible(groupDisplayEl);
        }
        const groupBy = ObjectUtil.get(record, this.groupBy);
        const r = this.groupByRenderer(groupBy, record, this);
        if (r) {
            if (typeof r === "string") {
                groupDisplayEl.innerHTML = r;
            }
            else if (r instanceof Component) {
                r.render(groupDisplayEl);
            }
            else if (r instanceof Promise) {
                r.then((s) => {
                    if (!s) {
                        return;
                    }
                    if (s instanceof Component) {
                        s.render(groupDisplayEl);
                    }
                    else {
                        groupDisplayEl.innerHTML = s;
                    }
                });
            }
            else {
                console.warn("Renderer returned invalid value: ", r);
            }
        }
    }
    renderRow(record, storeIndex) {
        const row = E(this.itemTag)
            .cls('+data')
            .attr('tabindex', '0');
        if (this.draggable) {
            row.draggable = true;
        }
        const r = this.renderer(record, row, this, storeIndex);
        if (typeof r === "string") {
            row.innerHTML = r;
        }
        else if (Array.isArray(r)) {
            r.forEach(c => {
                c.parent = this;
                c.render(row);
            });
        } // else NO-OP renderder will be responsible for appending html to the row @see Table
        if (this.rowSelection && this.rowSelection.isSelected(record)) {
            row.classList.add("selected");
        }
        return row;
    }
    onMouseEvent(e, type) {
        const row = this.findRowByEvent(e), index = row ? this.getRowElements().indexOf(row) : -1;
        if (index !== -1) {
            this.fire(type, this, index, row, e);
        }
    }
    findRowByEvent(e) {
        return e.target.closest(".data");
    }
}
/**
 * Shorthand function to create {@link Table}
 *
 * @param config
 */
export const list = (config) => createComponent(new List(config.store, config.renderer), config);
//# sourceMappingURL=List.js.map