import { Listener, Observable, ObservableEventMap, ObservableListenerOpts } from "./Observable.js";
import { Component } from "./Component.js";
type SortableDragEvent = DragEvent & {
    target: HTMLElement;
    /**
     * Set drag component to show instead of a translucent clone of the draggable
     *
     * ```
     * ev.setDragComponent(comp({cls: "card pad", html: this.rowSelect.selected.length + " selected rows"}))
     * ```
     *
     * @link https://developer.mozilla.org/en-US/docs/Web/API/DataTransfer/setDragImage
     * @param comp
     */
    setDragComponent: (comp: Component) => void;
};
export interface SortableEventMap<Type> extends ObservableEventMap<Type> {
    /**
     * Fires when the items are sorted
     *
     * @param dropComp The component the element was dropped on
     * @param fromIndex Move from index
     * @param toIndex To index
     * @param droppedOn Dropped on the toIndex or moved to this index
     * @param fromComp The component the element was dragged from if "group" is used to drop to other components
     * @param dragDataSet Arbitrary data set to dragData in dragstart event
     */
    sort: (dropComp: Type, toIndex: number, fromIndex: number, droppedOn: boolean, fromComp: Component, dragDataSet: Record<string, any>) => void;
    /**
     * Fires when the items are dragged over this sortable.
     *
     * Return false to disallow dropping
     *
     * @param dropComp The component the element was dropped on
     * @param fromIndex Move from index
     * @param toIndex To index
     * @param droppedOn Dropped on the toIndex or moved to this index
     * @param fromComp The component the element was dragged from if "group" is used to drop to other components
     */
    dropallowed: (dropComp: Type, toIndex: number, fromIndex: number, droppedOn: boolean, fromComp: Component, dragDataSet: Record<string, any>) => void;
    /**
     * Fires when dragging starts.
     *
     * You can set a component to show instead of a copy of the drag source:
     *
     * ev.setDragComponent(comp({cls: "card pad", html: this.rowSelect.selected.length + " selected rows"}))
     *
     * @param sortable
     * @param ev
     */
    dragstart: (sortable: Sortable<any>, ev: SortableDragEvent, dragData: DragData) => void;
    dragend: (sortable: Sortable<any>, ev: DragEvent, dragData: DragData) => void;
}
export interface Sortable<Type extends Component> {
    on<K extends keyof SortableEventMap<Type>, L extends Listener>(eventName: K, listener: Partial<SortableEventMap<Type>>[K], options?: ObservableListenerOpts): L;
    un<K extends keyof SortableEventMap<Type>>(eventName: K, listener: Partial<SortableEventMap<Type>>[K]): boolean;
    fire<K extends keyof SortableEventMap<Type>>(eventName: K, ...args: Parameters<SortableEventMap<any>[K]>): boolean;
}
type DROP_POSITION = "before" | "after" | "on";
type DragData = {
    dragSrc: HTMLElement | undefined;
    pos: DROP_POSITION;
    overEl: HTMLElement | undefined;
    fromIndex: number;
    toIndex: number;
    group: string;
    fromComponent: Component;
    toComponent: Component;
    dataSet: Record<string, any>;
};
/**
 * Enables sorting of child elements inside a container
 */
export declare class Sortable<Type extends Component> extends Observable {
    readonly component: Type;
    private sortableChildSelector;
    /**
     * Only allow drag and drop to Sortable's from the same group
     */
    group: string | undefined;
    /**
     * Allow dropping on child nodes
     */
    dropOn: boolean;
    /**
     * Allow sorting
     */
    dropBetween: boolean;
    /**
     * Set to true when items are stacked horizontally
     */
    horizontal: boolean;
    /**
     * Move the DOM nodes. Often GOUI components have underlying data stores that need to be updated. The components update
     * the UI already when the data in their store changes.
     */
    updateDom: boolean;
    private _gap;
    private static _dropPin;
    private static getDropPin;
    private static _dragImg;
    private static getDragImg;
    /**
     * Find the index of the item in the list of sortables
     *
     * @param sortableItem
     * @private
     */
    private findIndex;
    /**
     * Constructor
     *
     * @param component The container component where the sortables are inside
     * @param sortableChildSelector CSS selector to find the sortables. For example a css class ".sortable-item".
     */
    constructor(component: Type, sortableChildSelector: string);
    private getDragPos;
    private dropAllowed;
    /**
     * Calculates the gap between two rows to place the drop pin between them
     *
     * @private
     */
    private gap;
    private endDrag;
    private findSortables;
}
export {};
//# sourceMappingURL=Sortable.d.ts.map