import { Observable } from "./Observable.js";
import { comp, Component } from "./Component.js";
import { root } from "./Root.js";
const dragData = {
    dragSrc: undefined,
    pos: "before",
    overEl: undefined,
    fromIndex: -1,
    toIndex: -1,
    group: "",
    fromComponent: root,
    toComponent: root,
    dataSet: {}
};
/**
 * Enables sorting of child elements inside a container
 */
export class Sortable extends Observable {
    static getDropPin() {
        if (Sortable._dropPin == undefined) {
            Sortable._dropPin = comp({
                cls: "drop-pin",
                hidden: true
            });
            root.items.add(Sortable._dropPin);
        }
        return Sortable._dropPin;
    }
    static getDragImg() {
        if (Sortable._dragImg == undefined) {
            Sortable._dragImg = comp({
                cls: "drag-img"
            });
            root.items.add(Sortable._dragImg);
        }
        return Sortable._dragImg;
    }
    /**
     * Find the index of the item in the list of sortables
     *
     * @param sortableItem
     * @private
     */
    findIndex(sortableItem) {
        let index = 0, curr = sortableItem;
        while ((curr = curr.previousElementSibling)) {
            if (curr.matches(this.sortableChildSelector)) {
                index++;
            }
        }
        return index;
    }
    /**
     * Constructor
     *
     * @param component The container component where the sortables are inside
     * @param sortableChildSelector CSS selector to find the sortables. For example a css class ".sortable-item".
     */
    constructor(component, sortableChildSelector) {
        super();
        this.component = component;
        this.sortableChildSelector = sortableChildSelector;
        /**
         * Allow dropping on child nodes
         */
        this.dropOn = false;
        /**
         * Allow sorting
         */
        this.dropBetween = true;
        /**
         * Set to true when items are stacked horizontally
         */
        this.horizontal = false;
        /**
         * Move the DOM nodes. Often GOUI components have underlying data stores that need to be updated. The components update
         * the UI already when the data in their store changes.
         */
        this.updateDom = false;
        component.el.on("dragstart", (e) => {
            if (!this.group) {
                this.group = "sortable-" + Component.uniqueID();
            }
            // does this item belong to us? Not sure if this is efficient. Perhaps two sortables on the same el is madness.
            if (this.findSortables().indexOf(e.target) === -1) {
                return;
            }
            // had to add this class because otherwise dragleave fires immediately on child nodes: https://stackoverflow.com/questions/7110353/html5-dragleave-fired-when-hovering-a-child-element
            root.el.cls("+dragging");
            dragData.group = this.group;
            dragData.dragSrc = e.target;
            dragData.fromComponent = this.component;
            dragData.toComponent = this.component;
            dragData.fromIndex = this.findIndex(e.target);
            e.dataTransfer.setData('text/plain', 'goui');
            e.dataTransfer.effectAllowed = "copyMove";
            e.target.classList.add("drag-src");
            const ev = Object.assign(e, { setDragComponent: function (comp) {
                    Sortable.getDragImg().items.replace(comp);
                    this.dataTransfer.setDragImage(comp.el, 0, 0);
                }
            });
            this.fire("dragstart", this, ev, dragData);
        });
        component.el.on("drop", (e) => {
            e.preventDefault();
            this.endDrag(e);
        });
        component.el.on("dragend", (e) => {
            e.preventDefault();
            this.endDrag(e);
        });
        component.el.on("dragover", (e) => {
            if (dragData.group != this.group) {
                return;
            }
            if (!this.dropOn && !this.dropBetween) {
                return;
            }
            dragData.overEl = e.target.closest(this.sortableChildSelector);
            if (!this.dropAllowed()) {
                return;
            }
            const dropPin = Sortable.getDropPin();
            dragData.toComponent = component;
            if (dragData.overEl) {
                const rect = dragData.overEl.getBoundingClientRect();
                dragData.pos = this.getDragPos(rect, e);
                switch (dragData.pos) {
                    case "before":
                        e.preventDefault();
                        e.stopPropagation();
                        e.dataTransfer.dropEffect = "move";
                        dropPin.hidden = false;
                        if (this.horizontal) {
                            dropPin.el.style.top = rect.y + "px"; // (rect.y - this.gap(dragData.overEl)) + "px";
                            dropPin.el.style.left = (rect.x - this.gap(dragData.overEl)) + "px";
                            dropPin.el.style.height = rect.height + "px";
                        }
                        else {
                            dropPin.el.style.top = (rect.y - this.gap(dragData.overEl)) + "px";
                            dropPin.el.style.left = rect.x + "px";
                            dropPin.el.style.width = rect.width + "px";
                        }
                        break;
                    case "on":
                        if (!this.dropOn) {
                            return;
                        }
                        e.dataTransfer.dropEffect = "copy";
                        dropPin.hidden = true;
                        e.preventDefault();
                        e.stopPropagation();
                        break;
                    case "after":
                        e.preventDefault();
                        e.stopPropagation();
                        e.dataTransfer.dropEffect = "move";
                        dropPin.hidden = false;
                        if (this.horizontal) {
                            dropPin.el.style.top = rect.y + "px";
                            dropPin.el.style.left = (rect.x + rect.width + this.gap(dragData.overEl)) + "px";
                            dropPin.el.style.height = rect.height + "px";
                            dropPin.el.style.width = "";
                        }
                        else {
                            dropPin.el.style.top = (rect.y + rect.height + this.gap(dragData.overEl)) + "px";
                            dropPin.el.style.left = rect.x + "px";
                            dropPin.el.style.width = rect.width + "px";
                            dropPin.el.style.height = "";
                        }
                        break;
                }
            }
            else {
                dropPin.hidden = true;
                e.dataTransfer.dropEffect = "copy";
            }
        });
    }
    getDragPos(rect, e) {
        let pos;
        if (this.dropBetween) {
            if (this.horizontal) {
                const betweenZone = this.dropOn ? 10 : rect.width / 2;
                pos = "before";
                if (e.x > rect.x + betweenZone) {
                    pos = "on";
                }
                if (e.x > rect.x + rect.width - betweenZone) {
                    pos = "after";
                }
            }
            else {
                const betweenZone = this.dropOn ? 10 : rect.height / 2;
                pos = "before";
                if (e.y > rect.y + betweenZone) {
                    pos = "on";
                }
                if (e.y > rect.y + rect.height - betweenZone) {
                    pos = "after";
                }
            }
        }
        else {
            pos = "on";
        }
        return pos;
    }
    dropAllowed() {
        if (dragData.overEl) {
            dragData.toIndex = this.findIndex(dragData.overEl);
            if (dragData.pos == "after") {
                dragData.toIndex++;
            }
        }
        else {
            dragData.toIndex = 0;
        }
        return this.fire("dropallowed", dragData.toComponent, dragData.toIndex, dragData.fromIndex, dragData.pos == "on", dragData.fromComponent, dragData.dataSet);
    }
    /**
     * Calculates the gap between two rows to place the drop pin between them
     *
     * @private
     */
    gap(item) {
        if (this._gap !== undefined) {
            return this._gap;
        }
        let item2 = item.nextElementSibling, r1, r2;
        if (item2 == null) {
            item2 = item.previousElementSibling;
            if (item2 === null) {
                return 0;
            }
            r1 = item2.getBoundingClientRect();
            r2 = item.getBoundingClientRect();
        }
        else {
            r1 = item.getBoundingClientRect();
            r2 = item2.getBoundingClientRect();
        }
        this._gap = this.horizontal
            ? (r1.x + r1.width - r2.x) / 2
            : (r1.y + r1.height - r2.y) / 2;
        return this._gap;
    }
    endDrag(e) {
        Sortable.getDropPin().hidden = true;
        if (dragData.group != this.group) {
            //not our item
            return;
        }
        if (!dragData.dragSrc) {
            this.fire("dragend", this, e, dragData);
            return;
        }
        dragData.dragSrc.classList.remove("drag-src");
        root.el.cls("-dragging");
        if ((!this.dropOn && !this.dropBetween) || !this.dropAllowed()) {
            dragData.dragSrc = undefined;
            this.fire("dragend", this, e, dragData);
            return;
        }
        if (dragData.overEl) {
            dragData.toIndex = this.findIndex(dragData.overEl);
            if (dragData.pos == "after") {
                dragData.toIndex++;
            }
            if (this.updateDom) {
                switch (dragData.pos) {
                    case "before":
                        dragData.overEl.parentNode.insertBefore(dragData.dragSrc, dragData.overEl);
                        break;
                    case "on":
                        dragData.dragSrc.parentNode.removeChild(dragData.dragSrc);
                        break;
                    case "after":
                        dragData.overEl.parentNode.insertBefore(dragData.dragSrc, dragData.overEl.nextSibling);
                        break;
                }
            }
        }
        dragData.dragSrc = undefined;
        this.fire("sort", dragData.toComponent, dragData.toIndex, dragData.fromIndex, dragData.pos == "on", dragData.fromComponent, dragData.dataSet);
        this.fire("dragend", this, e, dragData);
    }
    findSortables() {
        return Array.from(this.component.el.querySelectorAll(this.sortableChildSelector));
    }
}
//# sourceMappingURL=Sortable.js.map