import { Component, createComponent } from "./Component";
import { root } from "./Root";
/**
 * Tooltip component
 *
 * @example TIP: For long HTML tooltips use the render event:
 * ```
 * tooltip({
 * 	listeners: {'render': (me) => {me.html = 'Long text'}},
 * 	target: someDiv
 * });
 * ```
 */
export class ToolTip extends Component {
    constructor() {
        super("menu");
        this.renderTo = root.el;
        /**
         * Tooltip shows with a delay in ms
         */
        this.delay = 300;
        this.closeDelay = 100;
        this.timeout = undefined;
        this.baseCls = "goui-dropdown";
    }
    set target(targetEl) {
        this.targetEl = targetEl;
        targetEl.on('mouseenter', e => {
            this.timeout = setTimeout(() => {
                this.open(e);
            }, this.delay);
        }).on('mouseleave', _e => {
            this.close();
        }).on('contextmenu', _e => {
            this.close();
        });
        this.el.on("mouseenter", () => {
            // keep tooltip open if moving mouse into the tooltip
            if (this.closeTimeout) {
                clearTimeout(this.closeTimeout);
                this.closeTimeout = undefined;
            }
        }).on('mouseleave', (e) => {
            //close when leaving tooltip
            this.close();
        });
    }
    internalRender() {
        if (!this.observer) {
            this.observer = new MutationObserver((mutationsList) => {
                mutationsList.forEach((mutation) => {
                    mutation.removedNodes.forEach((removedNode) => {
                        if (removedNode === this.targetEl) {
                            this.observer.disconnect(); // Optional: stop observing after removal
                            this.remove();
                        }
                    });
                });
            });
        }
        this.observer.observe(this.targetEl.parentNode, { childList: true });
        return super.internalRender();
    }
    internalRemove() {
        this.observer?.disconnect();
        super.internalRemove();
    }
    open(e) {
        this.render();
        this.align({ top: e.y, bottom: e.y, left: e.x, right: e.x });
        const pad = Component.remToPx(16);
        this.constrainTo(window, {
            top: pad,
            left: pad,
            bottom: pad,
            right: pad
        });
    }
    close() {
        if (this.timeout) {
            clearTimeout(this.timeout);
            this.timeout = undefined;
        }
        //close delay to allow moving into tooltip
        this.closeTimeout = setTimeout(() => {
            this.remove();
        }, this.closeDelay);
    }
    align(rect) {
        let left = rect.right + 16;
        let top = rect.bottom + 8;
        const tipRect = this.el.getBoundingClientRect();
        if (left + tipRect.width > window.innerWidth) {
            left = rect.left - tipRect.width - 8;
        }
        if (top + tipRect.height > window.innerHeight) {
            top = rect.top - tipRect.height - 16;
        }
        this.el.style.left = `${left}px`;
        this.el.style.top = `${top}px`;
    }
}
/**
 * Creates as {@link ToolTip} Component
 *
 * @param config
 *
 * @example
 * ```
 * tooltip({
 * 	text: "Hi I am the tooltip",
 * 	target: cmp.el
 * });
 * ```
 */
export const tooltip = (config) => createComponent(new ToolTip(), config);
//# sourceMappingURL=ToolTip.js.map