import { ChipsField } from "./ChipsField.js";
import { menu } from "../menu/index.js";
import { btn } from "../Button.js";
import { createComponent } from "../Component.js";
import { FunctionUtil } from "../../util/index.js";
/**
 * Chips component that auto completes user input
 */
export class AutocompleteChips extends ChipsField {
    /**
     * Constructor
     *
     * @param list The drop down list or {@link Table}.
     * @param buffer
     */
    constructor(list, buffer = 300) {
        super();
        this.list = list;
        this.buffer = buffer;
        //disable create
        this.textInputToValue = async (text) => {
            return false;
        };
        this.menu = menu({
            cls: "goui-dropdown scroll",
            removeOnClose: false,
            height: 400
        }, list);
        this.menuButton = btn({
            icon: "expand_more",
            type: "button",
            handler: (button, ev) => {
                this.fire("autocomplete", this, "");
            },
            menu: this.menu
        });
        this.initList();
    }
    initList() {
        if (!this.list.rowSelection) {
            this.list.rowSelectionConfig = { multiSelect: false };
        }
        if (!this.list.rowSelection.multiSelect) {
            this.list.on("rowclick", () => {
                this.addSelected();
                this.menu.hide();
            });
            this.list.store.on("datachanged", () => {
                this.list.rowSelection.selectIndex(0);
            }, { buffer: 0 });
        }
        else {
            const syncSelection = () => {
                this.list.rowSelection.clear();
                this.list.store.data.forEach((record, index) => {
                    if (this.isPickerRecordInValue(record)) {
                        this.list.rowSelection.add(record);
                    }
                });
            };
            this.menu.on("show", () => {
                syncSelection();
            }, { buffer: 0 });
            this.list.store.on("datachanged", () => {
                syncSelection();
            }, { buffer: 0 });
            this.menu.on("hide", () => {
                this.addSelected();
            });
        }
    }
    isPickerRecordInValue(record) {
        if (!this.value || !this.value.length) {
            return false;
        }
        if (!this.valuesToCompare) {
            this.valuesToCompare = this.value.map((i) => JSON.stringify(i));
            // clear cache after event loop
            setTimeout(() => {
                this.valuesToCompare = undefined;
            });
        }
        const v = JSON.stringify(this.pickerRecordToValue(this, record));
        return this.valuesToCompare.indexOf(v) > -1;
    }
    /**
     * Method that transforms a record from the TablePicker store to a value for this field.
     * This is not necessarily a text value. In conjunction with {@link valueToTextField()} this
     * could also be an ID of an object for example.
     *
     * @param field
     * @param record
     */
    pickerRecordToValue(field, record) {
        return record.id;
    }
    internalRender() {
        this.buttons = this.buttons || [];
        this.buttons.push(this.menuButton);
        const el = super.internalRender();
        this.menu.alignTo = this.wrap;
        this.menu.alignToInheritWidth = true;
        this.editor.el.addEventListener('input', FunctionUtil.buffer(this.buffer, this.onInput.bind(this)));
        return el;
    }
    onEditorKeyDown(ev) {
        super.onEditorKeyDown(ev);
        switch (ev.key) {
            case 'ArrowDown':
                ev.preventDefault();
                ev.stopPropagation();
                if (this.menuButton.menu.hidden) {
                    if (!this.list.store.loaded) {
                        this.fire("autocomplete", this, "");
                    }
                    this.menuButton.menu.show();
                }
                else if (this.list.rowSelection) {
                    this.list.rowSelection.selectNext();
                }
                break;
            case 'ArrowUp':
                ev.preventDefault();
                ev.stopPropagation();
                if (this.menuButton.menu.hidden) {
                    this.fire("autocomplete", this, "");
                    this.menuButton.menu.show();
                }
                else if (this.list.rowSelection) {
                    this.list.rowSelection.selectPrevious();
                }
                break;
            case 'Escape':
                if (!this.menu.hidden) {
                    this.menu.hide();
                    ev.preventDefault();
                    ev.stopPropagation();
                    this.focus();
                }
                break;
        }
    }
    onEnter(ev) {
        if (!this.list.rowSelection.getSelected().length) {
            super.onEnter(ev);
        }
        else {
            this.addSelected();
        }
        if (!this.menu.hidden) {
            ev.preventDefault();
            this.menu.hide();
        }
    }
    addSelected() {
        const newValues = this.list.rowSelection.getSelected().map((row) => {
            return this.pickerRecordToValue(this, row.record);
        });
        this.list.rowSelection.clear();
        this.editor.el.innerText = "";
        this.focus();
        // set value after focus as this will start tracking for the change event
        if (this.list.rowSelection.multiSelect) {
            this.value = newValues;
        }
        else {
            if (!this.value) {
                this.value = [];
            }
            this.value = this.value.concat(newValues);
        }
    }
    onInput(_ev) {
        this.menuButton.menu.show();
        this.fire("autocomplete", this, this.editor.el.innerText);
    }
}
/**
 * Shorthand function to create {@link AutocompleteChips}
 *
 * @param config
 */
export const autocompletechips = (config) => createComponent(new AutocompleteChips(config.list), config);
//# sourceMappingURL=AutocompleteChips.js.map