/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { FunctionUtil } from "../../util/index.js";
import { btn } from "../Button.js";
import { createComponent } from "../Component.js";
import { listpicker } from "../picker/index.js";
import { Menu, menu } from "../menu/index.js";
import { InputField } from "./InputField.js";
/**
 * Autocomplete field
 */
export class AutocompleteField extends InputField {
    /**
     *
     * @param list The table to use for suggestions
     * @param buffer Buffer typing in the input in ms
     */
    constructor(list, buffer = 300) {
        super();
        this.list = list;
        this.buffer = buffer;
        this.autocomplete = "off";
        this.baseCls = "goui-form-field text autocomplete";
        this.picker = listpicker({
            list: list
        });
        this.picker.on("select", (tablePicker, record) => {
            tablePicker.list.findAncestorByType(Menu).hide();
            this.focus();
            //set value after focus for change event
            this.value = this.pickerRecordToValue(this, record);
            this.fire('select', this, record);
        });
        this.menu = menu({
            height: 300,
            cls: "scroll",
            listeners: {
                hide: (menu) => {
                    if (menu.rendered) {
                        if (this.value == undefined) {
                            this.input.value = "";
                        }
                        const inputField = menu.findAncestorByType(InputField);
                        inputField.focus();
                    }
                }
            }
        }, this.picker);
        this.menuButton = btn({
            icon: "expand_more",
            type: "button",
            handler: () => {
                if (this.list.rowSelection) {
                    this.list.rowSelection.clear();
                }
                this.fire("autocomplete", this, "");
                this.input.focus();
            },
            listeners: {
                render: comp => {
                    comp.el.addEventListener("keydown", (ev) => {
                        switch (ev.key) {
                            case 'ArrowDown':
                                ev.preventDefault();
                                if (this.list.rowSelection) {
                                    this.list.rowSelection.clear();
                                }
                                this.fire("autocomplete", this, "");
                                this.menuButton.menu.show();
                                this.input.focus();
                                break;
                        }
                    });
                }
            },
            menu: this.menu
        });
        this.fireChangeOnBlur = true;
    }
    createInput() {
        const control = document.createElement("input");
        // select the text so users can type right away
        control.addEventListener("focus", function () {
            this.select();
        });
        if (this.invalidMsg) {
            this.applyInvalidMsg();
        }
        return control;
    }
    /**
     * Method that transforms a record from the TablePicker store to a value for this field.
     * This is not necessarily a text value. In conjunction with {@link valueToTextField()} this
     * could also be an ID of an object for example.
     */
    pickerRecordToValue(_field, record) {
        return record.id;
    }
    /**
     * This method transforms the value in to a text representation for the input field
     */
    async valueToTextField(_field, _value) {
        return "";
    }
    internalSetValue(v) {
        if (v == undefined) {
            return super.internalSetValue(v);
        }
        this.valueToTextField(this, v + "").then(textFieldValue => {
            if (v != this.value) {
                //Quick and dirty way to cancel this promise if the value has changed in the mean time
                return;
            }
            if (this.input) {
                super.internalSetValue(textFieldValue);
            }
            else {
                this.on("render", () => {
                    super.internalSetValue(textFieldValue);
                }, { once: true });
            }
        });
    }
    internalGetValue() {
        return this._value;
    }
    internalRender() {
        this.buttons = this.buttons || [];
        if (this.clearable) {
            this.buttons.push(btn({ icon: 'clear', handler: () => this.value = null }));
        }
        this.buttons.push(this.menuButton);
        const el = super.internalRender();
        this.menu.alignTo = this.wrap;
        this.menu.alignToInheritWidth = true;
        this.input.addEventListener('input', FunctionUtil.buffer(this.buffer, this.onInput.bind(this)));
        this.input.addEventListener('keydown', (ev) => {
            switch (ev.key) {
                case 'Enter':
                    if (!this.menu.hidden) {
                        ev.preventDefault();
                        this.picker.onSelect();
                    }
                    break;
                case 'ArrowDown':
                    ev.preventDefault();
                    ev.stopPropagation();
                    if (this.menuButton.menu.hidden) {
                        if (!this.list.store.loaded) {
                            this.fire("autocomplete", this, "");
                        }
                        this.menuButton.menu.show();
                    }
                    else if (this.list.rowSelection) {
                        this.list.rowSelection.selectNext();
                    }
                    break;
                case 'ArrowUp':
                    ev.preventDefault();
                    ev.stopPropagation();
                    if (this.menuButton.menu.hidden) {
                        this.fire("autocomplete", this, "");
                        this.menuButton.menu.show();
                    }
                    else if (this.list.rowSelection) {
                        this.list.rowSelection.selectPrevious();
                    }
                    break;
                case 'Escape':
                    if (!this.menu.hidden) {
                        this.menu.hide();
                        ev.preventDefault();
                        ev.stopPropagation();
                    }
                    break;
            }
        });
        return el;
    }
    onInput(_ev) {
        this.value = null;
        this.menuButton.menu.show();
        this.fire("autocomplete", this, this.input.value);
    }
    reset() {
        // By clearing the store the arrow down will reload the store and not select the last selected item.
        this.list.store.clear();
        super.reset();
    }
    clear() {
        // By clearing the store the arrow down will reload the store and not select the last selected item.
        this.list.store.clear();
        super.clear();
    }
}
/**
 * Shorthand function to create an {@link AutocompleteField}
 *
 * @param config
 */
export const autocomplete = (config) => createComponent(new AutocompleteField(config.list), config);
//# sourceMappingURL=AutocompleteField.js.map