import { AutocompleteField } from "./AutocompleteField.js";
import { datasourcestore, } from "../../data/index.js";
import { column, table } from "../table/index.js";
import { createComponent } from "../Component.js";
import { Format } from "../../util/index.js";
import { t } from "../../Translate";
export const ComboBoxDefaultRenderer = (field, r) => r ? r[field.displayProperty] : t("Not found");
/**
 * Combo box
 *
 * An extension of the Autocomplete field that simplifies the creation of a simple combobox with a
 * valueProperty and displayProperty.
 */
export class ComboBox extends AutocompleteField {
    /**
     * Constructor
     * @param dataSource
     * @param displayProperty
     * @param valueProperty
     * @param renderer
     * @param storeConfig
     * @param tableConfig
     * @param selectFirst Selects the first record on render
     */
    constructor(dataSource, displayProperty = "name", valueProperty = "id", renderer = ComboBoxDefaultRenderer, storeConfig = {
        queryParams: {
            limit: 50
        }
    }, tableConfig, selectFirst = false) {
        storeConfig.dataSource = dataSource;
        const dropDownTable = table(Object.assign({
            headers: false,
            fitParent: true,
            store: datasourcestore(storeConfig),
            columns: [
                column({
                    id: displayProperty,
                    resizable: true,
                    width: 312,
                    sortable: true,
                    renderer: (_columnValue, record) => {
                        return Format.escapeHTML(renderer(this, record));
                    }
                })
            ]
        }, tableConfig));
        super(dropDownTable);
        this.dataSource = dataSource;
        this.displayProperty = displayProperty;
        this.valueProperty = valueProperty;
        this.renderer = renderer;
        this.selectFirst = selectFirst;
        /**
         * When autocompleting from the datasource this filter name will be used.
         */
        this.filterName = "text";
        // load next data on scroll
        this.picker.list.store.addScrollLoader(this.menu.el);
        this.picker.list.store.queryParams.limit = 50;
        this.on("autocomplete", async (_field, input) => {
            this.list.store.queryParams.position = 0;
            if (!this.list.store.queryParams.filter) {
                this.list.store.queryParams.filter = {};
            }
            const filterName = (this.filterName ?? this.displayProperty);
            this.list.store.setFilter(filterName, { [filterName]: input });
            this.list.store.setFilter("combo", this.filter);
            await this.list.store.load();
        });
        if (selectFirst) {
            this.on('render', this.selectFirstRecord, { bind: this });
        }
    }
    async selectFirstRecord() {
        //auto select first entry if no value set
        if (!this.value) {
            const oldLimit = this.list.store.queryParams.limit;
            this.list.store.queryParams.limit = 1;
            const records = await this.list.store.load();
            this.list.store.queryParams.limit = oldLimit;
            //recheck cmp.value because window may have been loaded simultaneously.
            if (!this.value && records.length) {
                this.value = records[0].id;
                this.trackReset();
            }
        }
    }
    pickerRecordToValue(_field, record) {
        return record[this.valueProperty];
    }
    async valueToTextField(_field, value) {
        if (value === "") {
            return "";
        }
        const record = await this.dataSource.single(value);
        return this.renderer(_field, record);
    }
}
/**
 * Shorthand function to create {@link ComboBox}
 *
 * @link https://goui.io/#form/Select
 * @param config
 */
export const combobox = (config) => createComponent(new ComboBox(config.dataSource, config.displayProperty ?? "name", config.valueProperty ?? "id", config.renderer ?? ComboBoxDefaultRenderer, config.storeConfig ?? { queryParams: { limit: 50 } }, config.tableConfig ?? {}, config.selectFirst ?? false), config);
//# sourceMappingURL=ComboBox.js.map