/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Form } from "./Form.js";
import { t } from "../../Translate.js";
import { createComponent } from "../Component.js";
import { Window } from "../Window.js";
import { Format } from "../../util/index";
export class DataSourceForm extends Form {
    constructor(dataSource) {
        super();
        this.dataSource = dataSource;
        this.handler = async (form1) => {
            try {
                let data, v = this.currentId ? this.modified : this.value;
                if (this.fire('beforesave', this, v) === false) {
                    return;
                }
                //save because it changes when this.currentId is set
                const isNew = this.isNew;
                if (!isNew) {
                    data = await this.dataSource.update(this.currentId, v);
                }
                else {
                    data = await this.dataSource.create(v);
                    this.currentId = data.id;
                }
                if (data) {
                    this.fire('save', this, data, isNew);
                }
            }
            catch (e) {
                if (e.type == "invalidProperties") {
                    this.handleServerValidation(e);
                }
                if (this.fire('saveerror', this, e) === false) {
                    return;
                }
                console.error(t("Error"), e);
                if (e.type != "invalidProperties") {
                    void Window.error(e);
                }
            }
        };
    }
    handleServerValidation(error) {
        console.error(error);
        for (const propertyName in error.validationErrors) {
            const field = this.findField(propertyName);
            if (!field) {
                continue;
            }
            field.setInvalid(Format.escapeHTML(error.validationErrors[propertyName].description));
        }
        const invalid = this.findFirstInvalid();
        if (invalid) {
            invalid.focus();
        }
        this.setInvalid(t('You have errors in your form. The invalid fields are marked.'));
    }
    create(data) {
        this.clear();
        this.fire('load', this, data);
        if (data) {
            this.value = data;
            this.trackReset();
        }
    }
    /**
     * Clear the form and set it to the original unloaded state
     *
     * @param setValue When the form loads it's cleared but we don't need to set the value
     */
    clear(setValue = true) {
        super.clear(setValue);
        delete this.currentId;
    }
    get isNew() {
        return !this.currentId;
    }
    /**
     * Load an entity into the form
     *
     * @param id
     */
    async load(id) {
        this.mask();
        try {
            this.clear(false);
            this.currentId = id;
            let entity = await this.dataSource.single(id);
            if (!entity) {
                throw "Failed to load entity with id " + id;
            }
            this.fire('load', this, entity);
            this.value = entity;
        }
        catch (e) {
            console.error(t("Error"), e);
            if (this.fire('loaderror', this, e) !== false) {
                void Window.error(e);
            }
        }
        finally {
            this.unmask();
        }
    }
}
/**
 * Shorthand function to create {@link DataSourceForm}
 *
 * @param config
 * @param items
 */
export const datasourceform = (config, ...items) => createComponent(new DataSourceForm(config.dataSource), config, items);
//# sourceMappingURL=DataSourceForm.js.map