/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Field } from "./Field.js";
import { Component, createComponent } from "../Component.js";
import { DateTime, Format } from "../../util";
const defaultDisplayFieldRenderer = (v, field) => v ?? "";
/**
 * Display field
 *
 * Component used to display data only.
 *
 * Use cls: "pit" to style as an editable form field.
 *
 * A {@link Form} can be used to edit data but also to present data using display fields
 */
export class DisplayField extends Field {
    /**
     *
     * @param tagName Tagname defaults to ="label". If you pass another tag like h3 for example it will render a simple tag: <h3>{renderedValue}</h3>
     * @param renderer Renderer function for the value of the field
     */
    constructor(tagName = "label", renderer = defaultDisplayFieldRenderer) {
        super(tagName);
        this.tagName = tagName;
        this.renderer = renderer;
        this.baseCls = 'goui-display-field';
        /**
         * Escape value HTML
         *
         * {@link Format.escapeHTML}
         */
        this.escapeValue = true;
        /**
         * Hide this field when the value is empty
         */
        this.hideWhenEmpty = true;
        this.renderTagOnly = false;
        if (tagName != "label") {
            this.renderTagOnly = true;
        }
    }
    createControl() {
        return document.createElement("div");
    }
    internalSetValue(v) {
        // if(this.control) {
        const setFn = (str) => {
            if (str instanceof Component) {
                if (!this.renderTagOnly) {
                    this.control.innerHTML = "";
                    str.render(this.control);
                }
                else {
                    this.el.innerHTML = "";
                    str.render(this.el);
                }
                this.on("remove", () => {
                    str.remove();
                });
            }
            else {
                if (this.escapeValue && (typeof str == 'string')) {
                    str = str.htmlEncode();
                }
                if (!this.renderTagOnly) {
                    this.control.innerHTML = str + "";
                }
                else {
                    this.el.innerHTML = str + "";
                }
                if (this.hideWhenEmpty) {
                    this.hidden = !str;
                }
            }
        }, str = this.renderer(v, this);
        str instanceof Promise ? str.then(setFn) : setFn(str);
        // }
    }
    renderControl() {
        if (!this.renderTagOnly) {
            super.renderControl();
        }
    }
}
/**
 * Shortcut function to create a {@link DisplayField}
 *
 * @param config
 * @param items
 */
export const displayfield = (config, ...items) => createComponent(new DisplayField(config?.tagName ?? "label", config?.renderer ?? defaultDisplayFieldRenderer), config, items);
/**
 * Create display field with date icon and renderer
 *
 * @param config
 * @param items
 */
export const displaydatefield = (config, ...items) => {
    if (!config.icon)
        config.icon = "today";
    if (!config.renderer)
        config.renderer = (v) => v ? (new DateTime(v)).format(Format.dateFormat) : "";
    return createComponent(new DisplayField(config?.tagName ?? "label", config?.renderer ?? defaultDisplayFieldRenderer), config, items);
};
//# sourceMappingURL=DisplayField.js.map