/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Config, Listener, Observable, ObservableEventMap, ObservableListenerOpts } from "../Observable.js";
import { List } from "../List.js";
import { Store, StoreRecord, storeRecordType } from "../../data";
export interface RowSelectEventMap<Type extends Observable, StoreType extends Store = Store, RecordType extends StoreRecord = storeRecordType<StoreType>> extends ObservableEventMap<Type> {
    /**
     * Fires when selection changes.
     * This function is buffered. So when adding/removing selected items or when holding arrow on keyboard it
     * will only fire once at key up to prevent flooding the server with requests.
     *
     * @example get store record in list config
     * ```
     * rowSelection: {
     * 				multiSelect: true,
     * 				listeners: {
     * 					selectionchange: (rowSelect) => {
     * 						if(rowSelect.getSelected().length == 1) {
     * 							const table = rowSelect.getTable();
     * 							const record = table.getStore().getRecordAt(rowSelect.getSelected()[0]);
     * 						}
     * 					}
     * 				}
     * 			}
     * ```
     * @param rowSelect
     */
    selectionchange: (rowSelect: Type, selected: SelectedRow<StoreType, RecordType>[]) => void;
    /**
     * Fires when a row is selected
     * @param rowSelect
     * @param storeIndex
     */
    beforerowselect: (rowSelect: Type, row: SelectedRow<StoreType, RecordType>, append: boolean) => false | void;
    /**
     * Fires when a row is deselected
     *
     * @param rowSelect
     * @param storeIndex
     */
    beforerowdeselect: (rowSelect: Type, row: SelectedRow<StoreType, RecordType>) => false | void;
    /**
     * Fires when a row is selected
     * @param rowSelect
     * @param storeIndex
     */
    rowselect: (rowSelect: Type, row: SelectedRow<StoreType, RecordType>, append: boolean) => void;
    /**
     * Fires when a row is deselected
     *
     * @param rowSelect
     * @param storeIndex
     */
    rowdeselect: (rowSelect: Type, row: SelectedRow<StoreType, RecordType>) => void;
}
export interface RowSelect<StoreType extends Store = Store, RecordType extends StoreRecord = storeRecordType<StoreType>> {
    on<K extends keyof RowSelectEventMap<this, StoreType, RecordType>, L extends Listener>(eventName: K, listener: Partial<RowSelectEventMap<this, StoreType, RecordType>>[K], options?: ObservableListenerOpts): L;
    fire<K extends keyof RowSelectEventMap<this, StoreType, RecordType>>(eventName: K, ...args: Parameters<RowSelectEventMap<any, StoreType, RecordType>[K]>): boolean;
}
export declare class SelectedRow<StoreType extends Store, RecordType extends StoreRecord = storeRecordType<StoreType>> {
    readonly store: StoreType;
    readonly record: RecordType;
    constructor(store: StoreType, record: RecordType);
    get storeIndex(): number;
    get id(): any;
}
/**
 * Row selection model
 */
export declare class RowSelect<StoreType extends Store = Store, RecordType extends StoreRecord = storeRecordType<StoreType>> extends Observable {
    readonly list: List;
    private readonly selected;
    /**
     * Last selected index used for multi selection with shift
     * @private
     */
    lastIndex: number;
    private shiftStartIndex?;
    /**
     * Enable multi selection
     */
    multiSelect: boolean;
    private hasKeyUpListener;
    private readonly fireSelectionChange;
    /**
     * Constructor
     *
     * @param list
     */
    constructor(list: List);
    getSelected(): SelectedRow<StoreType, RecordType>[];
    private _listHasCheckbox?;
    private listHasCheckbox;
    /**
     * Clear the selection
     */
    clear(): void;
    /**
     * Select all
     */
    selectAll(): void;
    selectNext(): number;
    private incrementSelect;
    selectIndex(index: number): void;
    selectPrevious(): number;
    /**
     * Check if a record is selected
     * @param record
     */
    isSelected(record: RecordType): boolean;
    private getSelectedIndex;
    /**
     * When shift or ctrl, meta is used then do this on mousedown
     *
     * @param list
     * @param index
     * @param e
     * @private
     */
    private onRowMouseDown;
    /**
     * Click clears the selection handling this in click allows selections to be dragged
     * @param list
     * @param index
     * @param e
     * @private
     */
    private onRowClick;
    add(record: RecordType, append?: boolean): void;
    /**
     * Remove an index from the selection
     *
     * @param record
     * @param silent Don't fire events
     */
    remove(record: RecordType, silent?: boolean): void;
    toggle(record: RecordType): void;
    private onKeyDown;
}
export type RowSelectConfig<StoreType extends Store = Store> = Config<RowSelect<StoreType>, RowSelectEventMap<RowSelect, StoreType>, "list">;
/**
 * Shorthand function to create {@link RowSelect}
 *
 * @param config
 */
export declare const rowselect: <StoreType extends Store = Store<StoreRecord>>(config: RowSelectConfig<StoreType>) => RowSelect<StoreType, storeRecordType<StoreType>>;
//# sourceMappingURL=RowSelect.d.ts.map