/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Observable } from "../Observable.js";
import { FunctionUtil } from "../../util";
import { createComponent } from "../Component.js";
import { Table } from "./Table.js";
import { CheckboxSelectColumn } from "./TableColumns.js";
export class SelectedRow {
    constructor(store, record) {
        this.store = store;
        this.record = record;
    }
    get storeIndex() {
        return this.store.findIndex((r => r == this.record || (r[this.store.idField] != undefined && r[this.store.idField] == this.id)));
    }
    get id() {
        return this.record[this.store.idField];
    }
}
/**
 * Row selection model
 */
export class RowSelect extends Observable {
    /**
     * Constructor
     *
     * @param list
     */
    constructor(list) {
        super();
        this.list = list;
        this.selected = [];
        /**
         * Last selected index used for multi selection with shift
         * @private
         */
        this.lastIndex = -1;
        /**
         * Enable multi selection
         */
        this.multiSelect = true;
        this.hasKeyUpListener = false;
        this.list.el.cls('+rowselect');
        this.list.el.addEventListener("keydown", (e) => {
            this.onKeyDown(e);
        });
        this.list.on('rowmousedown', (table, index, row, e) => {
            this.onRowMouseDown(table, index, e);
        });
        this.list.on('rowclick', (table, index, row, e) => {
            this.onRowClick(table, index, e);
        });
        const fireSelectionChange = () => {
            this.fire('selectionchange', this, this.selected);
        };
        //buffer selection change so it doesn't fire many changes if rowSelection.add is called in a loop.
        this.fireSelectionChange = FunctionUtil.buffer(0, fireSelectionChange);
    }
    getSelected() {
        return this.selected;
    }
    listHasCheckbox() {
        if (this._listHasCheckbox !== undefined) {
            return this._listHasCheckbox;
        }
        if (!(this.list instanceof Table)) {
            this._listHasCheckbox = false;
        }
        else {
            this._listHasCheckbox = this.list.columns.findIndex((c) => {
                return c instanceof CheckboxSelectColumn;
            }) > -1;
        }
        return this._listHasCheckbox;
    }
    /**
     * Clear the selection
     */
    clear() {
        while (this.selected.length) {
            this.remove(this.selected[0].record);
        }
        this.lastIndex = -1;
    }
    /**
     * Select all
     */
    selectAll() {
        const store = this.list.store;
        for (let i = 0, c = store.count(); i < c; i++) {
            this.add(store.get(i));
        }
    }
    selectNext() {
        return this.incrementSelect(1);
    }
    incrementSelect(inc = 1) {
        const index = this.lastIndex + inc;
        if (index < 0 || index > this.list.store.count() - 1) {
            return this.lastIndex;
        }
        this.selectIndex(index);
        this.lastIndex = index;
        return index;
    }
    selectIndex(index) {
        this.clear();
        const r = this.list.store.get(index);
        if (r) {
            this.add(r);
        }
    }
    selectPrevious() {
        return this.incrementSelect(-1);
    }
    /**
     * Check if a record is selected
     * @param record
     */
    isSelected(record) {
        return this.getSelectedIndex(record) > -1;
    }
    getSelectedIndex(record) {
        const idField = this.list.store.idField;
        const id = record[idField];
        for (let i = 0, l = this.selected.length; i < l; i++) {
            if ((id != undefined && this.selected[i].id == id) || this.selected[i].record == record) {
                return i;
            }
        }
        return -1;
    }
    /**
     * When shift or ctrl, meta is used then do this on mousedown
     *
     * @param list
     * @param index
     * @param e
     * @private
     */
    onRowMouseDown(list, index, e) {
        const store = list.store;
        // e.preventDefault();
        if (e.shiftKey && this.multiSelect) {
            const start = Math.min(index, this.lastIndex);
            const end = Math.max(index, this.lastIndex);
            console.log(list.store.data, start, end);
            for (let i = start; i <= end; i++) {
                const record = store.get(i);
                if (record) {
                    this.add(record, true);
                }
            }
        }
        else if ((e.ctrlKey || e.metaKey) && this.multiSelect) {
            const record = store.get(index);
            if (record) {
                this.toggle(record);
            }
        }
        else {
            if (e.ctrlKey || e.metaKey) {
                const record = store.get(index);
                if (record) {
                    this.toggle(record);
                }
                this.lastIndex = index;
                return;
            }
            else {
                //handled by click instead of mousedown
                return;
            }
        }
        this.lastIndex = index;
    }
    /**
     * Click clears the selection handling this in click allows selections to be dragged
     * @param list
     * @param index
     * @param e
     * @private
     */
    onRowClick(list, index, e) {
        if (e.shiftKey || e.ctrlKey || e.metaKey) {
            return;
        }
        this.clear();
        this.add(list.store.get(index));
        this.lastIndex = index;
    }
    add(record, append = false) {
        if (this.isSelected(record)) {
            return;
        }
        if (!this.multiSelect) {
            this.clear();
        }
        const selectedRow = new SelectedRow(this.list.store, record);
        if (this.fire("beforerowselect", this, selectedRow, append) === false) {
            return;
        }
        this.selected.push(selectedRow);
        this.fire('rowselect', this, selectedRow, append);
        this.fireSelectionChange();
        this.lastIndex = selectedRow.storeIndex;
    }
    /**
     * Remove an index from the selection
     *
     * @param record
     * @param silent Don't fire events
     */
    remove(record, silent = false) {
        const index = this.getSelectedIndex(record);
        if (index == -1) {
            return;
        }
        const selectedRow = this.selected[index];
        if (this.fire("beforerowdeselect", this, selectedRow) === false) {
            return;
        }
        this.selected.splice(index, 1);
        if (!silent) {
            if (selectedRow.storeIndex == -1) {
                //record is not in store anymore.
                return;
            }
            this.fire('rowdeselect', this, selectedRow);
            this.fireSelectionChange();
        }
    }
    toggle(record) {
        if (this.isSelected(record)) {
            if (this.multiSelect) {
                this.remove(record);
            }
            else {
                this.clear();
            }
        }
        else {
            if (!this.multiSelect) {
                this.clear();
            }
            this.add(record, true);
        }
    }
    onKeyDown(e) {
        if (e.key == "Shift") {
            this.shiftStartIndex = this.lastIndex;
        }
        if (e.key == " ") {
            e.preventDefault();
            e.stopPropagation();
            this.toggle(this.list.store.get(this.lastIndex));
            return;
        }
        console.log(e.key);
        if (e.key != "ArrowDown" && e.key != "ArrowUp") {
            return;
        }
        e.preventDefault();
        e.stopPropagation();
        let index = 0, change = false;
        if (e.key == "ArrowDown") {
            index = this.lastIndex + 1;
        }
        else if (e.key == "ArrowUp") {
            index = this.lastIndex - 1;
        }
        console.log(index, this.list.store.count());
        // check if index is out of bounds
        if (index < 0 || index > this.list.store.count() - 1) {
            return;
        }
        if (e.shiftKey && this.multiSelect) {
            if ((e.key == "ArrowDown" && index > this.shiftStartIndex) || (e.key == "ArrowUp" && index < this.shiftStartIndex)) {
                const record = this.list.store.get(index);
                this.add(record, true);
            }
            else {
                const record = this.list.store.get(this.lastIndex);
                this.remove(record);
            }
        }
        else {
            if (!this.listHasCheckbox()) {
                this.clear();
                this.add(this.list.store.get(index));
            }
            this.list.focusRow(index);
        }
        if (change && !this.hasKeyUpListener) {
            this.hasKeyUpListener = true;
            this.list.el.addEventListener('keyup', () => {
                this.fire('selectionchange', this, this.selected);
                this.hasKeyUpListener = false;
            }, { once: true });
        }
        this.lastIndex = index;
    }
}
/**
 * Shorthand function to create {@link RowSelect}
 *
 * @param config
 */
export const rowselect = (config) => createComponent(new RowSelect(config.list), config);
//# sourceMappingURL=RowSelect.js.map