/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Observable } from "../Observable.js";
import { createComponent } from "../Component.js";
import { Format } from "../../util/index.js";
import { checkbox } from "../form/index.js";
import { btn } from "../Button.js";
import { menu } from "../menu/index.js";
export class TableColumn extends Observable {
    /**
     *
     * The column ID. Also used for 'property'
     */
    constructor(id) {
        super();
        this.id = id;
        /**
         * Make the column resizable by the user
         */
        this.resizable = false;
        /**
         * Make it sortable by the user
         */
        this.sortable = false;
        /**
         * Text alignment
         */
        this.align = "left";
        /**
         * Hide the column. It can be enabled by the user via the context menu.
         */
        this.hidden = false;
        /**
         * Enable this column in the enabled columns menu
         */
        this.hidable = true;
        this.property = id;
    }
}
/**
 * Create a table column
 *
 * @param config
 */
export const column = (config) => createComponent(new TableColumn(config.id), config);
export class DateTimeColumn extends TableColumn {
    constructor() {
        super(...arguments);
        this.renderer = (date) => {
            return Format.smartDateTime(date);
        };
        this.align = "right";
        this.width = 190;
    }
}
/**
 * Create a column showing date and time
 * @param config
 */
export const datetimecolumn = (config) => createComponent(new DateTimeColumn(config.id), config);
export class DateColumn extends TableColumn {
    constructor() {
        super(...arguments);
        this.renderer = (date) => {
            return Format.smartDateTime(date, false);
        };
        this.align = "right";
        this.width = 128;
    }
}
/**
 * Create a column showing just a date
 *
 * @param config
 */
export const datecolumn = (config) => createComponent(new DateColumn(config.id), config);
export class BoolColumn extends TableColumn {
    constructor() {
        super(...arguments);
        this.renderer = (v) => {
            return v ? '<i class="icon">check</i>' : "";
        };
        this.align = "center";
        this.width = 64;
    }
}
/**
 * Create a column showing a boolean value as check or empty
 *
 * @param config
 */
export const boolcolumn = (config) => createComponent(new BoolColumn(config.id), config);
export class CheckboxColumn extends TableColumn {
    constructor(id) {
        super(id);
        this.renderer = (val, record, td, table, rowIndex, column) => {
            return checkbox({
                value: val,
                listeners: {
                    change: (field, newValue, oldValue) => {
                        record[column.property] = newValue;
                        this.fire("change", this, field, newValue, record, rowIndex);
                    },
                    render: (field) => {
                        table.el.addEventListener("keydown", (e) => {
                            if (e.key == " " && e.target == td.parentNode) {
                                field.value = !field.value;
                            }
                        });
                    },
                }
            });
        };
        this.cls = "checkbox-select-column";
    }
}
/**
 * Create a checkbox column
 *
 * @param config
 */
export const checkboxcolumn = (config) => createComponent(new CheckboxColumn(config && config.id ? config.id : "checkbox"), config);
export class CheckboxSelectColumn extends TableColumn {
    constructor(id = "checkboxselect") {
        super(id);
        this.headerRenderer = (col, headerEl, table) => {
            return checkbox({
                listeners: {
                    change: (field, newValue, oldValue) => {
                        if (newValue) {
                            table.rowSelection.selectAll();
                        }
                        else {
                            table.rowSelection.clear();
                        }
                    }
                }
            });
        };
        this.renderer = (val, record, td, table, rowIndex) => {
            // add to selection model if value is true
            if (val && table.rowSelection) {
                table.rowSelection.add(record);
            }
            return checkbox({
                value: val,
                listeners: {
                    render: (field) => {
                        field.el.addEventListener("mousedown", (ev) => {
                            ev.stopPropagation();
                        });
                        field.el.addEventListener("click", (ev) => {
                            ev.stopPropagation();
                        });
                        field.value = table.rowSelection.isSelected(record);
                        table.rowSelection.on("selectionchange", () => {
                            field.value = table.rowSelection.isSelected(record);
                        });
                    },
                    change: (field, newValue, oldValue) => {
                        if (newValue) {
                            table.rowSelection.add(record);
                        }
                        else {
                            table.rowSelection.remove(record);
                        }
                    }
                }
            });
        };
        this.hidable = false;
        this.sticky = true;
        this.cls = "checkbox-select-column";
    }
}
export const checkboxselectcolumn = (config) => createComponent(new CheckboxSelectColumn(config && config.id ? config.id : "checkboxselect"), config);
/**
 * Creates a menu button.
 *
 * All items will have a property dataSet.rowIndex and dataSet.table so you know which record has been clicked on.
 *
 * @param items
 */
export const menucolumn = (...items) => column({
    width: 48,
    id: "btn",
    renderer: (columnValue, record, td, table, rowIndex) => {
        items.forEach(i => {
            i.dataSet.table = table;
            i.dataSet.rowIndex = rowIndex;
        });
        return btn({
            icon: "more_vert",
            menu: menu({}, ...items)
        });
    }
});
//# sourceMappingURL=TableColumns.js.map