import { Table, TableColumn } from "./../table";
import { Config, Listener, ObservableListenerOpts } from "./../Observable.js";
import { ListEventMap } from "../List";
import { TreeStore } from "./TreeStore";
import { TreeRecord } from "./TreeRecord";
export interface TreeEventMap<Type> extends ListEventMap<Type> {
    /**
     * Fires before a node expands
     *
     * You can cancel the expand by returning false.
     *
     * Could be useful to populate the tree.
     *
     * @param tree The  tree
     * @param record The record of the expanding node
     * @param storeIndex The index of the record in the store
     */
    beforeexpand: (tree: Type, record: TreeRecord, storeIndex: number) => void | false;
    /**
     * Fires when a node expands
     *
     * @param tree The  tree
     * @param record The record of the expanding node
     * @param storeIndex The index of the record in the store
     */
    expand: (tree: Type, record: TreeRecord, storeIndex: number) => void;
    /**
     * Fires when a node collapses
     *
     * @param tree The main tree
     * @param record The record of the collapsing node
     * @param storeIndex The index of the record in the store
     */
    beforecollapse: (tree: Type, record: TreeRecord, storeIndex: number) => void | false;
    /**
     * Fires when a node collapses
     *
     * @param tree The main tree
     * @param record The record of the collapsing node
     * @param storeIndex The index of the record in the store
     */
    collapse: (tree: Type, record: TreeRecord, storeIndex: number) => void;
    /**
     * Fires when a node collapses
     *
     * @param tree The main tree
     * @param record The record of the collapsing node
     * @param storeIndex The index of the record in the store
     */
    checkchange: (tree: Type, record: TreeRecord, storeIndex: number, checked: boolean) => void;
}
export interface Tree extends Table<TreeStore> {
    on<K extends keyof TreeEventMap<this>, L extends Listener>(eventName: K, listener: Partial<TreeEventMap<this>>[K], options?: ObservableListenerOpts): L;
    un<K extends keyof TreeEventMap<this>>(eventName: K, listener: Partial<TreeEventMap<this>>[K]): boolean;
    fire<K extends keyof TreeEventMap<this>>(eventName: K, ...args: Parameters<TreeEventMap<any>[K]>): boolean;
}
export type NodeProvider = (record: TreeRecord | undefined, store: TreeStore) => TreeRecord[] | Promise<TreeRecord[]>;
/**
 * Tree component
 */
export declare class Tree extends Table<TreeStore> {
    protected nodeProvider: NodeProvider;
    constructor(nodeProvider: NodeProvider, columns?: TableColumn[]);
    protected renderRow(record: any, storeIndex: number): HTMLElement;
    /**
     * Expand a tree node
     *
     * @param record
     */
    expand(record: any): Promise<void>;
    /**
     * Collapse a tree node
     *
     * @param record
     */
    collapse(record: any): void;
    private internalExpand;
    private internalCollapse;
    /**
     * Expands a node after 1s when dragging something over it
     *
     * @param row
     * @param record
     * @param storeIndex
     * @private
     */
    private setupExpandOnDragOver;
}
/**
 * Generator function for a {@link Tree} component
 *
 * @param config
 */
export declare const tree: (config: Config<Tree, TreeEventMap<Tree>> & {
    nodeProvider: NodeProvider;
}) => Tree;
//# sourceMappingURL=Tree.d.ts.map