import { Table } from "./../table";
import { createComponent } from ".././Component";
import { TreeStore } from "./TreeStore";
import { TreeColumn } from "./TreeColumn";
/**
 * Tree component
 */
export class Tree extends Table {
    constructor(nodeProvider, columns = [new TreeColumn("text")]) {
        super(new TreeStore(nodeProvider), columns);
        this.nodeProvider = nodeProvider;
        this.headers = columns.length > 1;
        this.baseCls = this.baseCls + " tree";
        this.cls = "no-row-lines";
        this.on("render", async () => {
            void this.store.load();
        });
    }
    renderRow(record, storeIndex) {
        let row = super.renderRow(record, storeIndex);
        row.getElementsByClassName("caret")[0].on("click", (e) => {
            const node = row.getElementsByClassName("node")[0];
            const isExpanded = node.has(".expanded");
            if (isExpanded) {
                this.internalCollapse(record, storeIndex, node);
            }
            else {
                void this.internalExpand(record, storeIndex, node);
            }
            e.preventDefault();
            e.stopPropagation();
        });
        this.setupExpandOnDragOver(row, record, storeIndex);
        return row;
    }
    /**
     * Expand a tree node
     *
     * @param record
     */
    expand(record) {
        const index = this.store.findIndex(i => i == record);
        const row = this.getRowElements()[index];
        return this.internalExpand(record, index, row);
    }
    /**
     * Collapse a tree node
     *
     * @param record
     */
    collapse(record) {
        const index = this.store.findIndex(i => i == record);
        const row = this.getRowElements()[index];
        return this.internalCollapse(record, index, row);
    }
    async internalExpand(record, storeIndex, node) {
        if (this.fire("beforeexpand", this, record, storeIndex) === false) {
            return;
        }
        await this.store.expand(record);
        node.classList.add("expanded");
        this.fire("expand", this, record, storeIndex);
    }
    internalCollapse(record, storeIndex, node) {
        if (this.fire("beforecollapse", this, record, storeIndex) === false) {
            return;
        }
        this.store.collapse(record);
        node.classList.remove("expanded");
        this.fire("collapse", this, record, storeIndex);
    }
    /**
     * Expands a node after 1s when dragging something over it
     *
     * @param row
     * @param record
     * @param storeIndex
     * @private
     */
    setupExpandOnDragOver(row, record, storeIndex) {
        if (!this.dropOn && !this.dropBetween) {
            return;
        }
        let dragOverTimeout;
        row.addEventListener("dragenter", (e) => {
            if (dragOverTimeout) {
                clearTimeout(dragOverTimeout);
            }
            // this first setTimeout is needed otherwise dragleave will fire immediately. The .dragging class is assed
            // by sortable to the GOUI root el to set pointer-event = none on the children https://stackoverflow.com/questions/7110353/html5-dragleave-fired-when-hovering-a-child-element
            setTimeout(() => {
                //expand tree node if dragging over for 1 second
                dragOverTimeout = setTimeout(() => {
                    const node = row.getElementsByClassName("node")[0];
                    if (!node.has(".expanded")) {
                        void this.internalExpand(record, storeIndex, node);
                    }
                }, 1000);
            });
        });
        row.addEventListener("dragleave", (e) => {
            if (dragOverTimeout) {
                clearTimeout(dragOverTimeout);
            }
        });
        this.on("drop", () => {
            if (dragOverTimeout) {
                clearTimeout(dragOverTimeout);
            }
        });
    }
}
/**
 * Generator function for a {@link Tree} component
 *
 * @param config
 */
export const tree = (config) => createComponent(new Tree(config.nodeProvider, config.columns ?? [new TreeColumn("text")]), config);
//# sourceMappingURL=Tree.js.map