/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Store } from "../data/Store.js";
import { ObjectUtil } from "../util/index.js";
import { createComponent } from "../component/index.js";
/**
 * DataSourceStore class
 *
 * Uses an {@link AbstractDataSource} to present data in the view.
 * @category Data
 */
export class DataSourceStore extends Store {
    constructor(dataSource) {
        super();
        this.dataSource = dataSource;
        this.queryParams = {};
        this.hasMore = false;
        /**
         * The server properties required
         */
        this.properties = [];
        /**
         * Reload when the datasource changes
         */
        this.monitorChanges = true;
        /**
         * Builds record from entity
         * @param entity
         */
        this.buildRecord = async (entity) => entity;
        this.listening = false;
        this.keepPosition = false;
        this.filters = {};
        // Always start listening for changes. If no component is bound then it will always stay in memory.
        this.listen();
    }
    /**
     * Reloads the store when the datasource changes
     *
     * @protected
     */
    onDSChange(DataSource, changes) {
        if (this.loaded && this.monitorChanges && !this.loading) {
            void this.reload();
        }
    }
    bindComponent(comp) {
        super.bindComponent(comp);
        this.listen();
    }
    listen() {
        if (!this.listening) {
            this.dataSource.on('change', this.onDSChange, { bind: this });
            this.listening = true;
        }
    }
    unbindComponent(comp) {
        super.unbindComponent(comp);
        if (!this.components.length) {
            this.dataSource.un("change", this.onDSChange);
            this.listening = false;
        }
    }
    async internalLoad(append = false) {
        const queryParams = structuredClone(this.queryParams);
        queryParams.sort = this.sort;
        if (queryParams.limit) {
            //to see if the remote has more data we query one ID more.
            queryParams.limit++;
        }
        const queryResponse = await this.dataSource.query(queryParams);
        if (this.queryParams.limit) {
            // check if the server has more data.
            this.hasMore = queryResponse.ids.length > this.queryParams.limit;
            if (this.hasMore) {
                queryResponse.ids.pop();
            }
        }
        let list;
        if (queryResponse.list) {
            list = queryResponse.list;
        }
        else {
            const getResponse = await this.dataSource.get(queryResponse.ids, this.properties);
            list = getResponse.list;
        }
        const entities = await this.fetchRelations(list), records = await Promise.all(entities.map(this.buildRecord));
        this.loadData(records, append);
        return records;
    }
    async fetchRelations(records) {
        if (!this.relations) {
            return records;
        }
        const promises = [];
        for (const relationName in this.relations) {
            const rel = this.relations[relationName];
            let ids;
            for (const record of records) {
                ids = ObjectUtil.get(record, rel.path);
                if (!ids) {
                    continue;
                }
                if (!Array.isArray(ids)) {
                    promises.push(rel.dataSource.single(ids).then((e) => {
                        if (e) {
                            record[relationName] = e;
                        }
                    }));
                }
                else {
                    const idToEntity = (id) => {
                        return rel.dataSource.single(id);
                    };
                    promises.push(Promise.all(ids.map(idToEntity)).then((entities) => {
                        record[relationName] = entities;
                    }));
                }
            }
        }
        return Promise.all(promises).then(() => records);
    }
    reload() {
        const limit = this.queryParams.limit, pos = this.queryParams.position;
        this.queryParams.position = 0;
        if (limit) {
            this.queryParams.limit = Math.max(limit, this.data.length);
        }
        this.keepPosition = true;
        const r = super.reload();
        this.keepPosition = false;
        this.queryParams.limit = limit;
        this.queryParams.position = pos;
        return r;
    }
    load(append = false) {
        if (!this.keepPosition) {
            this.queryParams.position = 0;
        }
        return super.load(append);
    }
    loadNext(append = false) {
        if (!this.queryParams.limit) {
            throw new Error("Limit must be set for pagination");
        }
        this.queryParams.position = this.queryParams.position || 0;
        this.queryParams.position += this.queryParams.limit;
        this.keepPosition = true;
        const r = this.load(append);
        this.keepPosition = false;
        return r;
    }
    loadPrevious() {
        if (!this.queryParams.limit) {
            throw new Error("Limit and position must be set!");
        }
        this.queryParams.position = this.queryParams.position || 0;
        this.queryParams.position -= this.queryParams.limit;
        return this.load(false);
    }
    hasNext() {
        return this.hasMore;
    }
    hasPrevious() {
        return this.queryParams.position > 0;
    }
    /**
     * Patch an existing filter with extra filter options
     *
     * @param ref
     * @param filter
     */
    patchFilter(ref, filter) {
        const f = this.getFilter(ref) ?? {};
        return this.setFilter(ref, Object.assign(f, filter));
    }
    /**
     * Set a filter for the {@link AbstractDataSource::query()} method
     *
     * @param ref A reference name so it can be replaced later by the component.
     * @param filter
     */
    setFilter(ref, filter) {
        if (filter === undefined) {
            delete this.filters[ref];
        }
        else {
            this.filters[ref] = filter;
        }
        const conditions = [];
        for (const k in this.filters) {
            conditions.push(this.filters[k]);
        }
        switch (conditions.length) {
            case 0:
                delete this.queryParams.filter;
                break;
            case 1:
                this.queryParams.filter = conditions[0];
                break;
            default:
                this.queryParams.filter = {
                    operator: "AND",
                    conditions: conditions
                };
                break;
        }
        return this;
    }
    clearFilter(...names) {
        names.forEach(n => this.setFilter(n, undefined));
    }
    getFilter(name) {
        return this.filters[name];
    }
}
/**
 * Shorthand function to create a {@link DataSourceStore}
 *
 * @param config
 */
export const datasourcestore = (config) => {
    let f;
    if (config.filters) {
        f = config.filters;
        delete config.filters;
    }
    const store = createComponent(new DataSourceStore(config.dataSource), config);
    if (f) {
        for (const filterName in f) {
            store.setFilter(filterName, f[filterName]);
        }
    }
    return store;
};
//# sourceMappingURL=DataSourceStore.js.map