import { AbstractDataSource } from "./AbstractDataSource.js";
/**
 * REST based data source
 *
 * You probably need to extend this class and implement
 *
 * @category Data
 */
export class RestDataSource extends AbstractDataSource {
    /**
     * Constructor
     *
     * @param uri The base URI of the REST API. for example https://groupoffice.com/api
     * @param id The Data source ID. Will be appended to the base uri above. for Example
     *  if the ID is "users" the uri will be: "https://groupoffice.com/api/users"
     * @param baseRequestOptions
     */
    constructor(uri, id, baseRequestOptions = {
        mode: "cors",
        method: "GET"
    }) {
        id = id || "restDataSource";
        super(id);
        this.uri = uri;
        this.baseRequestOptions = baseRequestOptions;
        this.limitParamName = "limit";
        this.positionParamName = "offset";
        this.filterParamName = "searchQuery";
        this.sortParamName = "orderColumn";
        this.sortDirectionParamName = "orderDirection";
        // if(uri.slice(-1) != "/") {
        // 	this.uri  = uri + "/";
        // }
        // We can't persist with REST stores as we can't sync data like with JMAP
        this.persist = false;
    }
    read(id, path = "", options = {}) {
        if (typeof id === "number") {
            path += "/" + id;
        }
        return this.request(path, options);
    }
    doRequest(path = "", options = {}) {
        return this.request(path, options);
    }
    async request(path = "", options = {}, queryParams) {
        const opts = structuredClone(this.baseRequestOptions);
        Object.assign(opts, options);
        // if path does not start with slash or uri does not end with slash
        if (path != '' && path.slice(0, 1) !== '/' && this.uri.slice(-1) !== '/') {
            path = "/" + path;
        }
        let url = this.uri + path;
        if (queryParams && queryParams.size) {
            url += '?' + queryParams;
        }
        const response = await fetch(url, opts);
        if (!response.ok) {
            throw await this.createErrorFromResponse(response);
        }
        return await response.json();
    }
    /**
     * Override to create JMAP style errors. Form handles for example:
     *
     * {type:"invalidProperties", validationErrors: {fieldName: {description: "some error description"}}}
     *
     * @param response
     * @protected
     */
    async createErrorFromResponse(response) {
        const body = await response.text();
        return new Error("Request error: " + response.status + " " + response.statusText + "\n\n" + body);
    }
    /**
     * Transforms the entity to a string to post to the server.
     *
     * @param data
     * @protected
     */
    dataToPostBody(data) {
        return JSON.stringify({ data: data });
    }
    entityFromServerResponse(data) {
        return data.data;
    }
    internalCommit(params) {
        const response = {
            updated: {},
            created: {},
            destroyed: []
        }, promises = [];
        for (let id in params.update) {
            promises.push(this.request(id, {
                method: "PUT",
                body: this.dataToPostBody(params.update[id])
            })
                .then((data) => {
                if (!response.updated) {
                    response.updated = {};
                }
                response.updated[id] = this.entityFromServerResponse(data);
            })
                .catch((reason) => {
                if (!response.notUpdated) {
                    response.notUpdated = {};
                }
                response.notUpdated[id] = reason;
            }));
        }
        for (let id in params.create) {
            promises.push(this.request("", {
                method: "POST",
                body: this.dataToPostBody(params.create[id])
            })
                .then((data) => {
                if (!response.created) {
                    response.created = {};
                }
                response.created[id] = this.entityFromServerResponse(data);
            })
                .catch((reason) => {
                if (!response.notCreated) {
                    response.notCreated = {};
                }
                response.notCreated[id] = reason;
            }));
        }
        for (let id of params.destroy) {
            promises.push(this.request(id + "", {
                method: "DELETE"
            })
                .then(() => {
                if (!response.destroyed) {
                    response.destroyed = [];
                }
                response.destroyed.push(id);
            })
                .catch((reason) => {
                if (!response.notDestroyed) {
                    response.notDestroyed = {};
                }
                response.notDestroyed[id] = reason;
            }));
        }
        return Promise.all(promises).then(() => {
            return response;
        });
    }
    async internalGet(ids, properties) {
        const promises = [];
        ids.forEach((id) => {
            promises.push(this.request(id + "").then((data) => {
                return this.entityFromServerResponse(data);
            }));
        });
        const users = await Promise.all(promises);
        const r = {
            list: []
        };
        users.forEach(u => {
            r.list.push(u);
        });
        return r;
    }
    queryParamsToUrlParams(params) {
        const uriParams = new URLSearchParams();
        if (params.limit)
            uriParams.set(this.limitParamName, params.limit.toString());
        if (params.position) {
            uriParams.set(this.positionParamName, params.position.toString());
        }
        if (params.sort && params.sort.length) {
            uriParams.set(this.sortParamName, params.sort[0].property);
            uriParams.set(this.sortDirectionParamName, params.sort[0].isAscending ? "ASC" : "DESC");
        }
        if (params.urlQueryParams) {
            for (const k in params.urlQueryParams) {
                uriParams.set(k, params.urlQueryParams[k]);
            }
        }
        return uriParams;
    }
    queryParamsToRequestOptions(params) {
        return undefined;
    }
    async internalQuery(params) {
        const response = await this.request("", this.queryParamsToRequestOptions(params), this.queryParamsToUrlParams(params));
        if (!Array.isArray(response.data)) {
            throw "Invalid query response";
        }
        // immediately add data so we don't have to fetch it when data is retrieved using {@link get()} or {@link single()}
        this.entityFromServerResponse(response);
        const ids = response.data.map((r) => r.id);
        return {
            ids: ids,
            list: response.data,
            total: response.total
        };
    }
    /**
     * No sync implemented for REST. States are ignored and data must always be refreshed.
     * @param state
     * @protected
     */
    async internalRemoteChanges(state) {
        return {};
    }
    async internalMerge(ids) {
        return Promise.reject("Not implemented");
    }
}
//# sourceMappingURL=RestDataSource.js.map