/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { createComponent } from "../component";
import { ArrayUtil, Collection } from "../util";
/**
 * Generic data store used by components
 *
 * @category Data
 */
export class Store extends Collection {
    constructor() {
        // private static stores: Record<string, Store> = {};
        //
        // /**
        //  * If the ID is given the store can be fetched with
        //  */
        // public id?: string;
        super(...arguments);
        this._loaded = false;
        /**
         * Sort the data on field and direction
         */
        this.sort = [];
        /**
         * Defines the field that holds the record ID
         */
        this.idField = "id";
        /**
         * Register of all components bound to this store using {@link bindStore()}
         * @private
         */
        this.components = [];
    }
    /**
     * Find record by ID
     *
     * @param id
     */
    findById(id) {
        return this.find((r => r[this.idField] == id));
    }
    findIndexById(id) {
        return this.findIndex((r => r[this.idField] == id));
    }
    /**
     * True when the store is loading
     */
    get loading() {
        return this._loading;
    }
    /**
     * True when data has been loaded at least once
     */
    get loaded() {
        return this._loaded;
    }
    /**
     * Load a set of records
     *
     * @param records
     * @param append
     */
    loadData(records, append = true) {
        append ? this.add(...records) : this.replace(...records);
        this._loaded = true;
        this.fire("load", this, records, append);
    }
    set data(records) {
        this.loadData(records);
    }
    get data() {
        return this.getArray();
    }
    /**
     * Binds as component to this store so it can update when the store changes
     * @param comp
     */
    bindComponent(comp) {
        this.components.push(comp);
        // handling remove and add per items allows a drag and drop action via store.remove and store.add
        this.on("remove", comp.onRecordRemove, { bind: comp });
        this.on("add", comp.onRecordAdd, { bind: comp });
        this.on("beforeload", comp.onBeforeStoreLoad, { bind: comp });
        this.on("load", comp.onStoreLoad, { bind: comp });
        this.on("loadexception", comp.onStoreLoadException, { bind: comp });
        comp.on("remove", () => {
            this.unbindComponent(comp);
        });
    }
    unbindComponent(comp) {
        this.un("remove", comp.onRecordRemove);
        this.un("add", comp.onRecordAdd);
        this.un("beforeload", comp.onBeforeStoreLoad);
        this.un("load", comp.onStoreLoad);
        this.un("loadexception", comp.onStoreLoadException);
        const i = this.components.indexOf(comp);
        this.components.splice(i, 1);
    }
    /**
     * Reload the data from the source
     */
    async reload() {
        this._loaded = false;
        return this.load();
    }
    clear() {
        this._loaded = false;
        return super.clear();
    }
    /**
     * Returns the loaded records. If append is true it only returns the new records.
     * Override this function for new store types.
     *
     * @param append
     * @protected
     */
    internalLoad(append) {
        // async so it works the same as data source stores. This way this.loading is true wghen loadData() is called
        return new Promise(resolve => setTimeout(() => {
            const sorted = ArrayUtil.multiSort(this.items, this.sort);
            this.loadData(sorted, append);
            resolve(this.items);
        }));
    }
    /**
     * Loads the data from source into the store
     *
     * @param append
     */
    load(append = false) {
        this.fire("beforeload", this, append);
        this._loading = this.internalLoad(append)
            .catch(reason => {
            console.error(reason);
            this.fire("loadexception", this, reason);
            throw reason;
        })
            .finally(() => {
            this._loading = undefined;
        });
        return this._loading;
    }
    /**
     * Load the next set of records when paging.
     * Doesn't do anything in the array store but can be implemented in async stores.
     */
    loadNext(append = false) {
        return Promise.resolve([]);
    }
    /**
     * Load the next set of records when paging.
     * Doesn't do anything in the array store but can be implemented in async stores.
     */
    loadPrevious() {
        return Promise.resolve([]);
    }
    hasNext() {
        return false;
    }
    hasPrevious() {
        return false;
    }
    /**
     * Load more data when this element is scrolled down
     *
     * @param el
     */
    addScrollLoader(el) {
        const onScroll = () => {
            const pixelsLeft = el.scrollHeight - el.scrollTop - el.offsetHeight;
            if (pixelsLeft < 100) {
                if (!this.loading && this.hasNext()) {
                    void this.loadNext(true);
                }
            }
        };
        el.addEventListener("scroll", onScroll, { passive: true });
        // this will fill the empty space on firt load.
        this.on("load", (store, records, append) => {
            // use set timeout otherwise this.loading is still true
            setTimeout(() => {
                onScroll();
            });
        });
        return onScroll;
    }
}
/**
 * Shorthand function to create a {@link Store}
 *
 * @param config
 */
export const store = (config) => createComponent(new Store(), config);
//# sourceMappingURL=Store.js.map