/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
/**
 * Utilities to operate on objects
 * @category Utility
 */
export class ObjectUtil {
    /**
     * Get a nested value by JSON pointer
     *
     * eg.
     * const obj = {
     *   foo : {
     *     bar: {
     *       test: 1
     *     }
     *   }
     * }
     *
     * Object.path(obj, 'foo/bar/test'); // 1
     *
     * @param obj
     * @param pointer JSON pointer
     *
     * @link https://www.rfc-editor.org/rfc/rfc6901
     *
     * @return The value from the path or undefined if not found
     */
    static get(obj, pointer) {
        const parts = ObjectUtil.explodePointer(pointer);
        let part, cur = obj; // = structuredClone(obj);
        while (part = parts.shift()) {
            if (!(part in cur) && parts.length) {
                throw "Invalid JSON pointer : " + pointer;
            }
            cur = cur[part];
        }
        return cur;
    }
    /**
     * Deep merge two key value objects
     * @param o1
     * @param o2
     */
    static merge(o1, o2) {
        if (!this.isObject(o2)) {
            return o2;
        }
        for (let key in o2) {
            if (key in o1 && this.isObject(o1[key])) {
                o1[key] = this.merge(o1[key], o2[key]);
            }
            else {
                o1[key] = o2[key];
            }
        }
        return o1;
    }
    /**
     * Clone an object
     *
     * @param source
     */
    static clone(source) {
        return structuredClone(source);
    }
    static explodePointer(path) {
        const parts = path.replace(/^\//, "").split('/');
        // ignore leading / as it is implicit
        for (let i = 0; i < parts.length; i++) {
            parts[i].replace('~1', '/')
                .replace('~0', '~');
        }
        return parts;
    }
    static internalPatch(doc, path, v) {
        const part = path.shift(), length = path.length;
        // CalendarItem.ts relies on this when patching participants. It can also patch participants when they are null.
        if (!doc) {
            doc = {}; // server will return null for empty maps
        }
        if (!(part in doc) && length > 0) {
            throw new Error('patching item in non-existing objects');
        }
        if (!length) {
            //remove property by setting null. See https://jmap.io/spec-core.html#set
            if (v === null || v === undefined) {
                delete doc[part];
            }
            else {
                doc[part] = v;
            }
        }
        else {
            doc[part] = ObjectUtil.internalPatch(doc[part], path, v);
        }
        return doc;
    }
    /**
     * Patch an object with a JSON Pointer object following rfc6901 standards
     *
     * @link https://www.rfc-editor.org/rfc/rfc6901
     * @param doc
     * @param patch
     */
    static patch(doc, patch) {
        for (const p in patch) {
            try {
                doc = ObjectUtil.internalPatch(doc, ObjectUtil.explodePointer(p), patch[p]);
            }
            catch (e) {
                console.warn("Error patching object: ", p, patch, doc);
            }
        }
        return doc;
    }
}
ObjectUtil.isObject = (obj) => {
    return Object.prototype.toString.call(obj) === '[object Object]';
};
//# sourceMappingURL=ObjectUtil.js.map